from enum import Enum
from typing import TYPE_CHECKING, Mapping, NamedTuple, Optional

import dagster._check as check
from dagster._core.definitions.metadata import (
    MetadataFieldSerializer,
    MetadataValue,
    normalize_metadata,
)
from dagster._serdes import whitelist_for_serdes
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster._utils.types import ExcInfo

if TYPE_CHECKING:
    from dagster._core.execution.context.system import StepExecutionContext


@whitelist_for_serdes(
    storage_field_names={"metadata": "metadata_entries"},
    field_serializers={"metadata": MetadataFieldSerializer},
)
class TypeCheckData(
    NamedTuple(
        "_TypeCheckData",
        [
            ("success", bool),
            ("label", str),
            ("description", Optional[str]),
            ("metadata", Mapping[str, MetadataValue]),
        ],
    )
):
    def __new__(cls, success, label, description=None, metadata=None):
        return super(TypeCheckData, cls).__new__(
            cls,
            success=check.bool_param(success, "success"),
            label=check.str_param(label, "label"),
            description=check.opt_str_param(description, "description"),
            metadata=normalize_metadata(
                check.opt_mapping_param(metadata, "metadata", key_type=str)
            ),
        )


@whitelist_for_serdes(
    storage_field_names={"metadata": "metadata_entries"},
    field_serializers={"metadata": MetadataFieldSerializer},
)
class UserFailureData(
    NamedTuple(
        "_UserFailureData",
        [
            ("label", str),
            ("description", Optional[str]),
            ("metadata", Mapping[str, MetadataValue]),
        ],
    )
):
    def __new__(cls, label, description=None, metadata=None):
        return super(UserFailureData, cls).__new__(
            cls,
            label=check.str_param(label, "label"),
            description=check.opt_str_param(description, "description"),
            metadata=normalize_metadata(
                check.opt_mapping_param(metadata, "metadata", key_type=str)
            ),
        )


@whitelist_for_serdes
class ErrorSource(Enum):
    # An error that occurs while executing framework code
    FRAMEWORK_ERROR = "FRAMEWORK_ERROR"
    # An error that occurs while executing user code
    USER_CODE_ERROR = "USER_CODE_ERROR"
    # An error occurred at an unexpected time
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"
    # Execution was interrupted
    INTERRUPT = "INTERRUPT"


@whitelist_for_serdes
class StepFailureData(
    NamedTuple(
        "_StepFailureData",
        [
            ("error", Optional[SerializableErrorInfo]),
            ("user_failure_data", Optional[UserFailureData]),
            ("error_source", ErrorSource),
        ],
    )
):
    def __new__(cls, error, user_failure_data, error_source=None):
        return super(StepFailureData, cls).__new__(
            cls,
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            user_failure_data=check.opt_inst_param(
                user_failure_data, "user_failure_data", UserFailureData
            ),
            error_source=check.opt_inst_param(
                error_source, "error_source", ErrorSource, default=ErrorSource.FRAMEWORK_ERROR
            ),
        )

    @property
    def error_display_string(self) -> str:
        """Creates a display string that hides framework frames if the error arose in user code."""
        if not self.error:
            return ""
        if self.error_source == ErrorSource.USER_CODE_ERROR:
            user_code_error = self.error.cause
            check.invariant(
                user_code_error,
                (
                    "User code error is missing cause. User code errors are expected to have a"
                    " causes, which are the errors thrown from user code."
                ),
            )
            return self.error.message.strip() + ":\n\n" + user_code_error.to_string()
        else:
            return self.error.to_string()


def step_failure_event_from_exc_info(
    step_context: "StepExecutionContext",
    exc_info: ExcInfo,
    user_failure_data: Optional[UserFailureData] = None,
    error_source: Optional[ErrorSource] = None,
):
    from dagster._core.events import DagsterEvent

    return DagsterEvent.step_failure_event(
        step_context=step_context,
        step_failure_data=StepFailureData(
            error=serializable_error_info_from_exc_info(exc_info),
            user_failure_data=user_failure_data,
            error_source=error_source,
        ),
    )


@whitelist_for_serdes
class StepRetryData(
    NamedTuple(
        "_StepRetryData",
        [("error", SerializableErrorInfo), ("seconds_to_wait", Optional[check.Numeric])],
    )
):
    def __new__(cls, error, seconds_to_wait=None):
        return super(StepRetryData, cls).__new__(
            cls,
            error=check.opt_inst_param(error, "error", SerializableErrorInfo),
            seconds_to_wait=check.opt_numeric_param(seconds_to_wait, "seconds_to_wait"),
        )


@whitelist_for_serdes
class StepSuccessData(NamedTuple("_StepSuccessData", [("duration_ms", float)])):
    def __new__(cls, duration_ms):
        return super(StepSuccessData, cls).__new__(
            cls, duration_ms=check.float_param(duration_ms, "duration_ms")
        )



"""wipe schedules table for 0.10.0.

Revision ID: b32a4f3036d2
Revises: 375e95bad550
Create Date: 2021-01-11 22:20:01.253271

"""
from alembic import op
from dagster._core.storage.migration.utils import get_currently_upgrading_instance, has_table
from sqlalchemy import inspect

# alembic magic breaks pylint


# revision identifiers, used by Alembic.
revision = "b32a4f3036d2"
down_revision = "375e95bad550"
branch_labels = None
depends_on = None


def upgrade():
    inspector = inspect(op.get_bind())

    if "postgresql" not in inspector.dialect.dialect_description:
        return

    instance = get_currently_upgrading_instance()
    if instance.scheduler:
        instance.scheduler.wipe(instance)

    #   No longer dropping the "schedules" table here, since
    #   the 0.10.0 migration checks for the presence of the "schedules"
    #   table during the migration from the "schedules" table to the "jobs"
    #   table - see create_0_10_0_schedule_ tables
    #   if has_table("schedules"):
    #       op.drop_table("schedules")

    if has_table("schedule_ticks"):
        op.drop_table("schedule_ticks")


def downgrade():
    pass



from contextlib import contextmanager
from typing import IO, Any, Generator, Mapping, Optional, Sequence

from typing_extensions import Self

import dagster._check as check
from dagster._core.storage.captured_log_manager import (
    CapturedLogContext,
    CapturedLogData,
    CapturedLogManager,
    CapturedLogMetadata,
    CapturedLogSubscription,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData

from .compute_log_manager import (
    MAX_BYTES_FILE_READ,
    ComputeIOType,
    ComputeLogFileData,
    ComputeLogManager,
)


class NoOpComputeLogManager(CapturedLogManager, ComputeLogManager, ConfigurableClass):
    def __init__(self, inst_data: Optional[ConfigurableClassData] = None):
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return NoOpComputeLogManager(inst_data=inst_data, **config_value)

    def enabled(self, _dagster_run, _step_key):
        return False

    def _watch_logs(self, dagster_run, step_key=None):
        pass

    def get_local_path(self, run_id: str, key: str, io_type: ComputeIOType) -> str:
        raise NotImplementedError()

    def is_watch_completed(self, run_id, key):
        return True

    def on_watch_start(self, dagster_run, step_key):
        pass

    def on_watch_finish(self, dagster_run, step_key):
        pass

    def download_url(self, run_id, key, io_type):
        return None

    def read_logs_file(self, run_id, key, io_type, cursor=0, max_bytes=MAX_BYTES_FILE_READ):
        return ComputeLogFileData(
            path=f"{key}.{io_type}", data=None, cursor=0, size=0, download_url=None
        )

    def on_subscribe(self, subscription):
        pass

    def on_unsubscribe(self, subscription):
        pass

    @contextmanager
    def capture_logs(self, log_key: Sequence[str]) -> Generator[CapturedLogContext, None, None]:
        yield CapturedLogContext(log_key=log_key)

    def is_capture_complete(self, log_key: Sequence[str]):
        return True

    @contextmanager
    def open_log_stream(
        self, log_key: Sequence[str], io_type: ComputeIOType
    ) -> Generator[Optional[IO], None, None]:
        yield None

    def get_log_data(
        self,
        log_key: Sequence[str],
        cursor: Optional[str] = None,
        max_bytes: Optional[int] = None,
    ) -> CapturedLogData:
        return CapturedLogData(log_key=log_key)

    def get_log_metadata(self, log_key: Sequence[str]) -> CapturedLogMetadata:
        return CapturedLogMetadata()

    def delete_logs(
        self, log_key: Optional[Sequence[str]] = None, prefix: Optional[Sequence[str]] = None
    ):
        pass

    def subscribe(
        self, log_key: Sequence[str], cursor: Optional[str] = None
    ) -> CapturedLogSubscription:
        return CapturedLogSubscription(self, log_key, cursor)

    def unsubscribe(self, subscription: CapturedLogSubscription):
        pass



from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Iterator,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TypeAlias

from dagster._config import (
    ALL_CONFIG_BUILTINS,
    ConfigType,
    Field,
    Permissive,
    Selector,
    Shape,
)
from dagster._config.pythonic_config import Config
from dagster._core.definitions.asset_layer import AssetLayer
from dagster._core.definitions.executor_definition import (
    ExecutorDefinition,
    execute_in_process_executor,
    in_process_executor,
)
from dagster._core.definitions.input import InputDefinition
from dagster._core.definitions.output import OutputDefinition
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.storage.output_manager import IOutputManagerDefinition
from dagster._core.storage.root_input_manager import IInputManagerDefinition
from dagster._core.types.dagster_type import ALL_RUNTIME_BUILTINS, construct_dagster_type_dictionary
from dagster._utils import check

from .configurable import ConfigurableDefinition
from .definition_config_schema import IDefinitionConfigSchema
from .dependency import DependencyStructure, GraphNode, Node, NodeHandle, NodeInput, OpNode
from .graph_definition import GraphDefinition
from .logger_definition import LoggerDefinition
from .op_definition import NodeDefinition, OpDefinition
from .resource_definition import ResourceDefinition

if TYPE_CHECKING:
    from .source_asset import SourceAsset


def define_resource_dictionary_cls(
    resource_defs: Mapping[str, ResourceDefinition],
    required_resources: AbstractSet[str],
) -> Shape:
    fields = {}
    for resource_name, resource_def in resource_defs.items():
        if resource_def.config_schema:
            is_required = None
            if resource_name not in required_resources:
                # explicitly make section not required if resource is not required
                # for the current mode
                is_required = False

            fields[resource_name] = def_config_field(
                resource_def,
                is_required=is_required,
                description=resource_def.description,
            )

    return Shape(fields=fields)


def remove_none_entries(ddict: Mapping[Any, Any]) -> dict:
    return {k: v for k, v in ddict.items() if v is not None}


def def_config_field(
    configurable_def: ConfigurableDefinition,
    is_required: Optional[bool] = None,
    description: Optional[str] = None,
) -> Field:
    return Field(
        Shape(
            {"config": configurable_def.config_field} if configurable_def.has_config_field else {}
        ),
        is_required=is_required,
        description=description,
    )


class RunConfigSchemaCreationData(NamedTuple):
    job_name: str
    nodes: Sequence[Node]
    graph_def: GraphDefinition
    dependency_structure: DependencyStructure
    executor_def: ExecutorDefinition
    resource_defs: Mapping[str, ResourceDefinition]
    logger_defs: Mapping[str, LoggerDefinition]
    ignored_nodes: Sequence[Node]
    required_resources: AbstractSet[str]
    direct_inputs: Mapping[str, Any]
    asset_layer: AssetLayer


def define_logger_dictionary_cls(creation_data: RunConfigSchemaCreationData) -> Shape:
    return Shape(
        {
            logger_name: def_config_field(logger_definition, is_required=False)
            for logger_name, logger_definition in creation_data.logger_defs.items()
        }
    )


def define_execution_field(executor_defs: Sequence[ExecutorDefinition], description: str) -> Field:
    default_in_process = False
    for executor_def in executor_defs:
        if executor_def == in_process_executor:
            default_in_process = True

    selector = selector_for_named_defs(executor_defs)

    if default_in_process:
        return Field(
            selector, default_value={in_process_executor.name: {}}, description=description
        )

    # If we are using the execute_in_process executor, then ignore all executor config.
    if len(executor_defs) == 1 and executor_defs[0] == execute_in_process_executor:
        return Field(Permissive(), is_required=False, default_value={}, description=description)

    return Field(selector, description=description)


def define_single_execution_field(executor_def: ExecutorDefinition, description: str) -> Field:
    return def_config_field(executor_def, description=description)


def define_run_config_schema_type(creation_data: RunConfigSchemaCreationData) -> ConfigType:
    execution_field = define_single_execution_field(
        creation_data.executor_def,
        "Configure how steps are executed within a run.",
    )

    top_level_node = GraphNode(
        name=creation_data.graph_def.name,
        definition=creation_data.graph_def,
        graph_definition=creation_data.graph_def,
    )

    fields = {
        "execution": execution_field,
        "loggers": Field(
            define_logger_dictionary_cls(creation_data),
            description="Configure how loggers emit messages within a run.",
        ),
        "resources": Field(
            define_resource_dictionary_cls(
                creation_data.resource_defs,
                creation_data.required_resources,
            ),
            description="Configure how shared resources are implemented within a run.",
        ),
        "inputs": get_inputs_field(
            node=top_level_node,
            handle=NodeHandle(top_level_node.name, parent=None),
            dependency_structure=creation_data.dependency_structure,
            resource_defs=creation_data.resource_defs,
            node_ignored=False,
            direct_inputs=creation_data.direct_inputs,
            input_source_assets={},
            asset_layer=creation_data.asset_layer,
        ),
    }

    if creation_data.graph_def.has_config_mapping:
        config_schema = cast(IDefinitionConfigSchema, creation_data.graph_def.config_schema)
        nodes_field = Field(
            {"config": config_schema.as_field()},
            description="Configure runtime parameters for ops or assets.",
        )
    else:
        nodes_field = Field(
            define_node_shape(
                nodes=creation_data.nodes,
                ignored_nodes=creation_data.ignored_nodes,
                dependency_structure=creation_data.dependency_structure,
                resource_defs=creation_data.resource_defs,
                asset_layer=creation_data.asset_layer,
                node_input_source_assets=creation_data.graph_def.node_input_source_assets,
            ),
            description="Configure runtime parameters for ops or assets.",
        )

    fields["ops"] = nodes_field

    return Shape(
        fields=remove_none_entries(fields),
    )


# Common pattern for a set of named definitions (e.g. executors)
# to build a selector so that one of them is selected
def selector_for_named_defs(named_defs) -> Selector:
    return Selector({named_def.name: def_config_field(named_def) for named_def in named_defs})


def get_inputs_field(
    node: Node,
    handle: NodeHandle,
    dependency_structure: DependencyStructure,
    resource_defs: Mapping[str, ResourceDefinition],
    node_ignored: bool,
    asset_layer: AssetLayer,
    input_source_assets: Mapping[str, "SourceAsset"],
    direct_inputs: Optional[Mapping[str, Any]] = None,
) -> Optional[Field]:
    direct_inputs = check.opt_mapping_param(direct_inputs, "direct_inputs")
    inputs_field_fields = {}
    for name, inp in node.definition.input_dict.items():
        inp_handle = NodeInput(node, inp)
        has_upstream = input_has_upstream(dependency_structure, inp_handle, node, name)
        if inp.input_manager_key:
            input_field = get_input_manager_input_field(node, inp, resource_defs)
        elif (
            # if you have asset definitions, input will be loaded from the source asset
            asset_layer.has_assets_defs
            and asset_layer.asset_key_for_input(handle, name)
            and not has_upstream
        ):
            input_field = None
        elif name in direct_inputs and not has_upstream:
            input_field = None
        elif name in input_source_assets and not has_upstream:
            input_field = None
        elif inp.root_manager_key and not has_upstream:
            input_field = get_input_manager_input_field(node, inp, resource_defs)
        elif inp.dagster_type.loader and not has_upstream:
            input_field = get_type_loader_input_field(node, name, inp)
        else:
            input_field = None

        if input_field:
            inputs_field_fields[name] = input_field

    if not inputs_field_fields:
        return None
    if node_ignored:
        return Field(
            Shape(inputs_field_fields),
            is_required=False,
            description=(
                "This op is not present in the current op selection, "
                "the input config values are allowed but ignored."
            ),
        )
    else:
        return Field(Shape(inputs_field_fields))


def input_has_upstream(
    dependency_structure: DependencyStructure,
    input_handle: NodeInput,
    node: Node,
    input_name: str,
) -> bool:
    return dependency_structure.has_deps(input_handle) or node.container_maps_input(input_name)


def get_input_manager_input_field(
    node: Node,
    input_def: InputDefinition,
    resource_defs: Mapping[str, ResourceDefinition],
) -> Optional[Field]:
    if input_def.root_manager_key:
        if input_def.root_manager_key not in resource_defs:
            raise DagsterInvalidDefinitionError(
                f"Input '{input_def.name}' for {node.describe_node()} requires root_manager_key"
                f" '{input_def.root_manager_key}', but no resource has been provided. Please"
                " include a resource definition for that key in the provided resource_defs."
            )

        root_manager = resource_defs[input_def.root_manager_key]
        if not isinstance(root_manager, IInputManagerDefinition):
            raise DagsterInvalidDefinitionError(
                f"Input '{input_def.name}' for {node.describe_node()} requires root_manager_key "
                f"'{input_def.root_manager_key}', but the resource definition provided is not an "
                "IInputManagerDefinition"
            )

        input_config_schema = root_manager.input_config_schema
        if input_config_schema:
            return input_config_schema.as_field()
        return None
    elif input_def.input_manager_key:
        if input_def.input_manager_key not in resource_defs:
            raise DagsterInvalidDefinitionError(
                f"Input '{input_def.name}' for {node.describe_node()} requires input_manager_key"
                f" '{input_def.input_manager_key}', but no resource has been provided. Please"
                " include a resource definition for that key in the provided resource_defs."
            )

        input_manager = resource_defs[input_def.input_manager_key]
        if not isinstance(input_manager, IInputManagerDefinition):
            raise DagsterInvalidDefinitionError(
                f"Input '{input_def.name}' for {node.describe_node()} requires input_manager_key "
                f"'{input_def.input_manager_key}', but the resource definition provided is not an "
                "IInputManagerDefinition"
            )

        input_config_schema = input_manager.input_config_schema
        if input_config_schema:
            return input_config_schema.as_field()
        return None

    return None


def get_type_loader_input_field(node: Node, input_name: str, input_def: InputDefinition) -> Field:
    loader = check.not_none(input_def.dagster_type.loader)
    return Field(
        loader.schema_type,
        is_required=(
            not node.definition.input_has_default(input_name) and not input_def.root_manager_key
        ),
    )


def get_outputs_field(
    node: Node,
    resource_defs: Mapping[str, ResourceDefinition],
) -> Optional[Field]:
    output_manager_fields = {}
    for name, output_def in node.definition.output_dict.items():
        output_manager_output_field = get_output_manager_output_field(
            node, output_def, resource_defs
        )
        if output_manager_output_field:
            output_manager_fields[name] = output_manager_output_field

    return Field(Shape(output_manager_fields)) if output_manager_fields else None


def get_output_manager_output_field(
    node: Node, output_def: OutputDefinition, resource_defs: Mapping[str, ResourceDefinition]
) -> Optional[ConfigType]:
    if output_def.io_manager_key not in resource_defs:
        raise DagsterInvalidDefinitionError(
            f'Output "{output_def.name}" for {node.describe_node()} requires io_manager_key '
            f'"{output_def.io_manager_key}", but no resource has been provided. Please include a '
            "resource definition for that key in the provided resource_defs."
        )
    if not isinstance(resource_defs[output_def.io_manager_key], IOutputManagerDefinition):
        raise DagsterInvalidDefinitionError(
            f'Output "{output_def.name}" for {node.describe_node()} requires io_manager_key '
            f'"{output_def.io_manager_key}", but the resource definition provided is not an '
            "IOutputManagerDefinition"
        )
    output_manager_def = resource_defs[output_def.io_manager_key]
    if (
        output_manager_def
        and isinstance(output_manager_def, IOutputManagerDefinition)
        and output_manager_def.output_config_schema
    ):
        return output_manager_def.output_config_schema.as_field()

    return None


def node_config_field(fields: Mapping[str, Optional[Field]], ignored: bool) -> Optional[Field]:
    trimmed_fields = remove_none_entries(fields)
    if trimmed_fields:
        if ignored:
            return Field(
                Shape(trimmed_fields),
                is_required=False,
                description=(
                    "This op is not present in the current op selection, "
                    "the config values are allowed but ignored."
                ),
            )
        else:
            return Field(Shape(trimmed_fields))
    else:
        return None


def construct_leaf_node_config(
    node: Node,
    handle: NodeHandle,
    dependency_structure: DependencyStructure,
    config_schema: Optional[IDefinitionConfigSchema],
    resource_defs: Mapping[str, ResourceDefinition],
    ignored: bool,
    asset_layer: AssetLayer,
    input_source_assets: Mapping[str, "SourceAsset"],
) -> Optional[Field]:
    return node_config_field(
        {
            "inputs": get_inputs_field(
                node,
                handle,
                dependency_structure,
                resource_defs,
                ignored,
                asset_layer,
                input_source_assets,
            ),
            "outputs": get_outputs_field(node, resource_defs),
            "config": config_schema.as_field() if config_schema else None,
        },
        ignored=ignored,
    )


def define_node_field(
    node: Node,
    handle: NodeHandle,
    dependency_structure: DependencyStructure,
    resource_defs: Mapping[str, ResourceDefinition],
    ignored: bool,
    asset_layer: AssetLayer,
    input_source_assets: Mapping[str, "SourceAsset"],
) -> Optional[Field]:
    # All nodes regardless of compositing status get the same inputs and outputs
    # config. The only thing the varies is on extra element of configuration
    # 1) Vanilla op definition: a 'config' key with the config_schema as the value
    # 2) Graph with field mapping: a 'config' key with the config_schema of
    #    the config mapping (via GraphDefinition#config_schema)
    # 3) Graph without field mapping: an 'ops' key with recursively defined
    #    ops dictionary
    # 4) `configured` graph with field mapping: a 'config' key with the config_schema that was
    #    provided when `configured` was called (via GraphDefinition#config_schema)

    assert isinstance(node, (OpNode, GraphNode)), f"Invalid node type: {type(node)}"

    if isinstance(node, OpNode):
        return construct_leaf_node_config(
            node,
            handle,
            dependency_structure,
            node.definition.config_schema,
            resource_defs,
            ignored,
            asset_layer,
            input_source_assets,
        )

    graph_def = node.definition

    if graph_def.has_config_mapping:
        # has_config_mapping covers cases 2 & 4 from above (only config mapped graphs can
        # be `configured`)...
        return construct_leaf_node_config(
            node,
            handle,
            dependency_structure,
            # ...and in both cases, the correct schema for 'config' key is exposed by this property:
            graph_def.config_schema,
            resource_defs,
            ignored,
            asset_layer,
            input_source_assets,
        )
        # This case omits an 'ops' key, thus if a graph is `configured` or has a field
        # mapping, the user cannot stub any config, inputs, or outputs for inner (child) nodes.
    else:
        fields = {
            "inputs": get_inputs_field(
                node,
                handle,
                dependency_structure,
                resource_defs,
                ignored,
                asset_layer,
                input_source_assets,
            ),
            "outputs": get_outputs_field(node, resource_defs),
            "ops": Field(
                define_node_shape(
                    nodes=graph_def.nodes,
                    ignored_nodes=None,
                    dependency_structure=graph_def.dependency_structure,
                    parent_handle=handle,
                    resource_defs=resource_defs,
                    asset_layer=asset_layer,
                    node_input_source_assets=graph_def.node_input_source_assets,
                )
            ),
        }

        return node_config_field(fields, ignored=ignored)


def define_node_shape(
    nodes: Sequence[Node],
    ignored_nodes: Optional[Sequence[Node]],
    dependency_structure: DependencyStructure,
    resource_defs: Mapping[str, ResourceDefinition],
    asset_layer: AssetLayer,
    node_input_source_assets: Mapping[str, Mapping[str, "SourceAsset"]],
    parent_handle: Optional[NodeHandle] = None,
) -> Shape:
    """Examples of what this method is used to generate the schema for:
    1.
        inputs: ...
        ops:
      >    op1: ...
      >    op2: ...

    2.
        inputs:
        ops:
          graph1: ...
            inputs: ...
            ops:
      >       op1: ...
      >       inner_graph: ...


    """
    ignored_nodes = check.opt_sequence_param(ignored_nodes, "ignored_nodes", of_type=Node)

    fields = {}
    for node in nodes:
        node_field = define_node_field(
            node,
            NodeHandle(node.name, parent_handle),
            dependency_structure,
            resource_defs,
            ignored=False,
            asset_layer=asset_layer,
            input_source_assets=node_input_source_assets.get(node.name, {}),
        )

        if node_field:
            fields[node.name] = node_field

    for node in ignored_nodes:
        node_field = define_node_field(
            node,
            NodeHandle(node.name, parent_handle),
            dependency_structure,
            resource_defs,
            ignored=True,
            asset_layer=asset_layer,
            input_source_assets=node_input_source_assets.get(node.name, {}),
        )
        if node_field:
            fields[node.name] = node_field

    return Shape(fields)


def iterate_node_def_config_types(node_def: NodeDefinition) -> Iterator[ConfigType]:
    if isinstance(node_def, OpDefinition):
        if node_def.has_config_field:
            yield from node_def.get_config_field().config_type.type_iterator()
    elif isinstance(node_def, GraphDefinition):
        for node in node_def.nodes:
            yield from iterate_node_def_config_types(node.definition)

    else:
        check.invariant(f"Unexpected NodeDefinition type {type(node_def)}")


def _gather_all_schemas(node_defs: Sequence[NodeDefinition]) -> Iterator[ConfigType]:
    dagster_types = construct_dagster_type_dictionary(node_defs)
    for dagster_type in list(dagster_types.values()) + list(ALL_RUNTIME_BUILTINS):
        if dagster_type.loader:
            yield from dagster_type.loader.schema_type.type_iterator()


def _gather_all_config_types(
    node_defs: Sequence[NodeDefinition], run_config_schema_type: ConfigType
) -> Iterator[ConfigType]:
    for node_def in node_defs:
        yield from iterate_node_def_config_types(node_def)

    yield from run_config_schema_type.type_iterator()


def construct_config_type_dictionary(
    node_defs: Sequence[NodeDefinition],
    run_config_schema_type: ConfigType,
) -> Tuple[Mapping[str, ConfigType], Mapping[str, ConfigType]]:
    type_dict_by_name = {t.given_name: t for t in ALL_CONFIG_BUILTINS if t.given_name}
    type_dict_by_key = {t.key: t for t in ALL_CONFIG_BUILTINS}
    all_types = list(_gather_all_config_types(node_defs, run_config_schema_type)) + list(
        _gather_all_schemas(node_defs)
    )

    for config_type in all_types:
        name = config_type.given_name
        if name and name in type_dict_by_name:
            if type(config_type) is not type(type_dict_by_name[name]):
                raise DagsterInvalidDefinitionError(
                    (
                        "Type names must be unique. You have constructed two different "
                        'instances of types with the same name "{name}".'
                    ).format(name=name)
                )
        elif name:
            type_dict_by_name[name] = config_type

        type_dict_by_key[config_type.key] = config_type

    return type_dict_by_name, type_dict_by_key


def _convert_config_classes_inner(configs: Any) -> Any:
    if not isinstance(configs, dict):
        return configs

    return {
        k: {"config": v._convert_to_config_dictionary()}  # noqa: SLF001
        if isinstance(v, Config)
        else _convert_config_classes_inner(v)
        for k, v in configs.items()
    }


def _convert_config_classes(configs: Dict[str, Any]) -> Dict[str, Any]:
    return _convert_config_classes_inner(configs)


class RunConfig:
    """Container for all the configuration that can be passed to a run. Accepts Pythonic definitions
    for op and asset config and resources and converts them under the hood to the appropriate config dictionaries.

    Example usage:

    .. code-block:: python

        class MyAssetConfig(Config):
            a_str: str

        @asset
        def my_asset(config: MyAssetConfig):
            assert config.a_str == "foo"

        materialize(
            [my_asset],
            run_config=RunConfig(
                ops={"my_asset": MyAssetConfig(a_str="foo")}
            )
        )

    """

    def __init__(
        self,
        ops: Optional[Dict[str, Any]] = None,
        resources: Optional[Dict[str, Any]] = None,
        loggers: Optional[Dict[str, Any]] = None,
        execution: Optional[Dict[str, Any]] = None,
    ):
        self.ops = check.opt_dict_param(ops, "ops")
        self.resources = check.opt_dict_param(resources, "resources")
        self.loggers = check.opt_dict_param(loggers, "loggers")
        self.execution = check.opt_dict_param(execution, "execution")

    def to_config_dict(self):
        return {
            "loggers": self.loggers,
            "resources": _convert_config_classes(self.resources),
            "ops": _convert_config_classes(self.ops),
            "execution": self.execution,
        }


CoercibleToRunConfig: TypeAlias = Union[Dict[str, Any], RunConfig]

T = TypeVar("T")


def convert_config_input(inp: Union[CoercibleToRunConfig, T]) -> Union[T, Mapping[str, Any]]:
    if isinstance(inp, RunConfig):
        return inp.to_config_dict()
    else:
        return inp



import warnings
from collections import defaultdict, namedtuple
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    NamedTuple,
    NoReturn,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvariantViolationError,
)
from dagster._utils import is_named_tuple_instance

from .config import ConfigMapping
from .dependency import (
    DependencyDefinition,
    DependencyMapping,
    DynamicCollectDependencyDefinition,
    IDependencyDefinition,
    MultiDependencyDefinition,
    NodeInvocation,
)
from .graph_definition import GraphDefinition
from .hook_definition import HookDefinition
from .inference import infer_output_props
from .input import InputDefinition, InputMapping
from .logger_definition import LoggerDefinition
from .node_definition import NodeDefinition
from .output import OutputDefinition, OutputMapping
from .policy import RetryPolicy
from .resource_definition import ResourceDefinition
from .utils import check_valid_name, validate_tags
from .version_strategy import VersionStrategy

if TYPE_CHECKING:
    from dagster._core.execution.execute_in_process_result import ExecuteInProcessResult
    from dagster._core.instance import DagsterInstance

    from .executor_definition import ExecutorDefinition
    from .job_definition import JobDefinition
    from .partition import PartitionedConfig, PartitionsDefinition
    from .source_asset import SourceAsset


_composition_stack: List["InProgressCompositionContext"] = []


class MappedInputPlaceholder:
    """Marker for holding places in fan-in lists where input mappings will feed."""


class InvokedNodeOutputHandle:
    """The return value for an output when invoking a node in a composition function."""

    node_name: str
    output_name: str
    node_type: str

    def __init__(self, node_name: str, output_name: str, node_type: str):
        self.node_name = check.str_param(node_name, "node_name")
        self.output_name = check.str_param(output_name, "output_name")
        self.node_type = check.str_param(node_type, "node_type")

    def __iter__(self) -> NoReturn:
        raise DagsterInvariantViolationError(
            'Attempted to iterate over an {cls}. This object represents the output "{out}" '
            'from the op/graph "{node}". Consider defining multiple Outs if you seek to pass '
            "different parts of this output to different op/graph.".format(
                cls=self.__class__.__name__, out=self.output_name, node=self.node_name
            )
        )

    def __getitem__(self, idx: object) -> NoReturn:
        raise DagsterInvariantViolationError(
            'Attempted to index in to an {cls}. This object represents the output "{out}" '
            "from the {described_node}. Consider defining multiple Outs if you seek to pass "
            "different parts of this output to different {node_type}s.".format(
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
                node_type=self.node_type,
            )
        )

    def describe_node(self) -> str:
        return f"{self.node_type} '{self.node_name}'"

    def alias(self, _) -> NoReturn:
        raise DagsterInvariantViolationError(
            "In {source} {name}, attempted to call alias method for {cls}. This object "
            'represents the output "{out}" from the already invoked {described_node}. Consider '
            "checking the location of parentheses.".format(
                source=current_context().source,
                name=current_context().name,
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
            )
        )

    def with_hooks(self, _) -> NoReturn:
        raise DagsterInvariantViolationError(
            "In {source} {name}, attempted to call hook method for {cls}. This object "
            'represents the output "{out}" from the already invoked {described_node}. Consider '
            "checking the location of parentheses.".format(
                source=current_context().source,
                name=current_context().name,
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
            )
        )


class InputMappingNode(NamedTuple):
    input_def: InputDefinition


class DynamicFanIn(NamedTuple):
    """Type to signify collecting over a dynamic output, output by collect() on a
    InvokedNodeDynamicOutputWrapper.
    """

    node_name: str
    output_name: str


InputSource: TypeAlias = Union[
    InvokedNodeOutputHandle,
    InputMappingNode,
    DynamicFanIn,
    "SourceAsset",
    List[Union[InvokedNodeOutputHandle, InputMappingNode]],
]


def _not_invoked_warning(
    node: "PendingNodeInvocation",
    context_source: str,
    context_name: str,
) -> None:
    warning_message = (
        "While in {context} context '{name}', received an uninvoked {node_type} '{node_name}'.\n"
    )
    if node.given_alias:
        warning_message += "'{node_name}' was aliased as '{given_alias}'.\n"
    if node.tags:
        warning_message += "Provided tags: {tags}.\n"
    if node.hook_defs:
        warning_message += "Provided hook definitions: {hooks}.\n"

    warning_message = warning_message.format(
        context=context_source,
        name=context_name,
        node_name=node.node_def.name,
        given_alias=node.given_alias,
        tags=node.tags,
        hooks=[hook.name for hook in node.hook_defs],
        node_type=node.node_def.node_type_str,
    )

    warnings.warn(warning_message.strip())


def enter_composition(name: str, source: str) -> None:
    _composition_stack.append(InProgressCompositionContext(name, source))


def exit_composition(
    output: Optional[Mapping[str, OutputMapping]] = None
) -> "CompleteCompositionContext":
    return _composition_stack.pop().complete(output)


def current_context() -> "InProgressCompositionContext":
    return _composition_stack[-1]


def is_in_composition() -> bool:
    return bool(_composition_stack)


def assert_in_composition(name: str, node_def: NodeDefinition) -> None:
    if len(_composition_stack) < 1:
        node_label = node_def.node_type_str
        correction = (
            f"Invoking {node_label}s is only valid in a function decorated with @job or @graph."
        )
        raise DagsterInvariantViolationError(
            f"Attempted to call {node_label} '{name}' outside of a composition function."
            f" {correction}"
        )


class InProgressCompositionContext:
    """This context captures invocations of nodes within a
    composition function such as @job or @graph.
    """

    name: str
    source: str
    _invocations: Dict[str, "InvokedNode"]
    _collisions: Dict[str, int]
    _pending_invocations: Dict[str, "PendingNodeInvocation"]

    def __init__(self, name: str, source: str):
        self.name = check.str_param(name, "name")
        self.source = check.str_param(source, "source")
        self._invocations = {}
        self._collisions = {}
        self._pending_invocations = {}

    def observe_invocation(
        self,
        given_alias: Optional[str],
        node_def: NodeDefinition,
        input_bindings: Mapping[str, InputSource],
        tags: Optional[Mapping[str, str]],
        hook_defs: Optional[AbstractSet[HookDefinition]],
        retry_policy: Optional[RetryPolicy],
    ) -> str:
        if given_alias is None:
            node_name = node_def.name
            self._pending_invocations.pop(node_name, None)
            if self._collisions.get(node_name):
                self._collisions[node_name] += 1
                node_name = "{node_name}_{n}".format(
                    node_name=node_name, n=self._collisions[node_name]
                )
            else:
                self._collisions[node_name] = 1
        else:
            node_name = given_alias
            self._pending_invocations.pop(node_name, None)

        if self._invocations.get(node_name):
            raise DagsterInvalidDefinitionError(
                "{source} {name} invoked the same node ({node_name}) twice without aliasing."
                .format(source=self.source, name=self.name, node_name=node_name)
            )

        self._invocations[node_name] = InvokedNode(
            node_name, node_def, input_bindings, tags, hook_defs, retry_policy
        )
        return node_name

    def add_pending_invocation(self, node: "PendingNodeInvocation") -> None:
        node_name = node.given_alias if node.given_alias else node.node_def.name
        self._pending_invocations[node_name] = node

    def complete(
        self, output: Optional[Mapping[str, OutputMapping]]
    ) -> "CompleteCompositionContext":
        return CompleteCompositionContext.create(
            self.name,
            self.source,
            self._invocations,
            check.opt_mapping_param(output, "output"),
            self._pending_invocations,
        )


class CompleteCompositionContext(NamedTuple):
    """The processed information from capturing node invocations during a composition function."""

    name: str
    node_defs: Sequence[NodeDefinition]
    dependencies: DependencyMapping[NodeInvocation]
    input_mappings: Sequence[InputMapping]
    output_mapping_dict: Mapping[str, OutputMapping]
    node_input_source_assets: Mapping[str, Mapping[str, "SourceAsset"]]

    @staticmethod
    def create(
        name: str,
        source: str,
        invocations: Mapping[str, "InvokedNode"],
        output_mapping_dict: Mapping[str, OutputMapping],
        pending_invocations: Mapping[str, "PendingNodeInvocation"],
    ) -> "CompleteCompositionContext":
        from .source_asset import SourceAsset

        dep_dict: Dict[NodeInvocation, Dict[str, IDependencyDefinition]] = {}
        node_def_dict: Dict[str, NodeDefinition] = {}
        input_mappings = []
        node_input_source_assets: Dict[str, Dict[str, "SourceAsset"]] = defaultdict(dict)

        for node in pending_invocations.values():
            _not_invoked_warning(node, source, name)

        for invocation in invocations.values():
            def_name = invocation.node_def.name
            if def_name in node_def_dict and node_def_dict[def_name] is not invocation.node_def:
                raise DagsterInvalidDefinitionError(
                    'Detected conflicting node definitions with the same name "{name}"'.format(
                        name=def_name
                    )
                )
            node_def_dict[def_name] = invocation.node_def

            deps: Dict[str, IDependencyDefinition] = {}
            for input_name, node in invocation.input_bindings.items():
                if isinstance(node, InvokedNodeOutputHandle):
                    deps[input_name] = DependencyDefinition(node.node_name, node.output_name)
                elif isinstance(node, InputMappingNode):
                    input_mappings.append(
                        node.input_def.mapping_to(invocation.node_name, input_name)
                    )
                elif isinstance(node, SourceAsset):
                    node_input_source_assets[invocation.node_name][input_name] = node
                elif isinstance(node, list):
                    entries: List[Union[DependencyDefinition, Type[MappedInputPlaceholder]]] = []
                    for idx, fanned_in_node in enumerate(node):
                        if isinstance(fanned_in_node, InvokedNodeOutputHandle):
                            entries.append(
                                DependencyDefinition(
                                    fanned_in_node.node_name, fanned_in_node.output_name
                                )
                            )
                        elif isinstance(fanned_in_node, InputMappingNode):
                            entries.append(MappedInputPlaceholder)
                            input_mappings.append(
                                fanned_in_node.input_def.mapping_to(
                                    invocation.node_name, input_name, idx
                                )
                            )
                        else:
                            check.invariant("Unexpected fanned in node received")

                    deps[input_name] = MultiDependencyDefinition(entries)
                elif isinstance(node, DynamicFanIn):
                    deps[input_name] = DynamicCollectDependencyDefinition(
                        node.node_name, node.output_name
                    )
                else:
                    check.failed(f"Unexpected input binding - got {node}")

            dep_dict[
                NodeInvocation(
                    invocation.node_def.name,
                    invocation.node_name,
                    tags=invocation.tags,
                    hook_defs=invocation.hook_defs,
                    retry_policy=invocation.retry_policy,
                )
            ] = deps

        return CompleteCompositionContext(
            name,
            list(node_def_dict.values()),
            dep_dict,
            input_mappings,
            output_mapping_dict,
            node_input_source_assets=node_input_source_assets,
        )


T_NodeDefinition = TypeVar("T_NodeDefinition", bound=NodeDefinition)


class PendingNodeInvocation(Generic[T_NodeDefinition]):
    """An intermediate object in composition that allows binding additional information before invoking.

    Users should not invoke this object directly.

    Examples:
        ..code-block:: python

            from dagster import graph, op

            @op
            def some_op():
                ...

            @graph
            def the_graph():
                # renamed_op is a PendingNodeInvocation object with an added
                # name attribute
                renamed_op = some_op.alias("new_name")
                renamed_op()

    """

    node_def: T_NodeDefinition
    given_alias: Optional[str]
    tags: Optional[Mapping[str, str]]
    hook_defs: AbstractSet[HookDefinition]
    retry_policy: Optional[RetryPolicy]

    def __init__(
        self,
        node_def: T_NodeDefinition,
        given_alias: Optional[str],
        tags: Optional[Mapping[str, str]],
        hook_defs: Optional[AbstractSet[HookDefinition]],
        retry_policy: Optional[RetryPolicy],
    ):
        self.node_def = check.inst_param(node_def, "node_def", NodeDefinition)
        self.given_alias = check.opt_str_param(given_alias, "given_alias")
        self.tags = check.opt_mapping_param(tags, "tags", key_type=str, value_type=str)
        self.hook_defs = check.opt_set_param(hook_defs, "hook_defs", HookDefinition)
        self.retry_policy = check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy)

        if self.given_alias is not None:
            check_valid_name(self.given_alias)

        if is_in_composition():
            current_context().add_pending_invocation(self)

    def __call__(self, *args, **kwargs) -> Any:
        from ..execution.context.invocation import UnboundOpExecutionContext
        from .decorators.op_decorator import DecoratedOpFunction
        from .op_invocation import op_invocation_result

        node_name = self.given_alias if self.given_alias else self.node_def.name

        # If PendingNodeInvocation is not within composition context, and underlying NodeDefinition
        # is an OpDefinition, then permit it to be invoked and executed like an OpDefinition.
        if not is_in_composition() and isinstance(self.node_def, OpDefinition):
            node_label = self.node_def.node_type_str
            # Pyright does not currently infer the typevar-bound type of self based on the above
            # `isinstance` check. We need the correct type for passing to `op_invocation_result`.
            _self = cast(PendingNodeInvocation[OpDefinition], self)
            if not isinstance(self.node_def.compute_fn, DecoratedOpFunction):
                raise DagsterInvalidInvocationError(
                    f"Attemped to invoke {node_label} that was not constructed using the"
                    f" `@{node_label}` decorator. Only {node_label}s constructed using the"
                    f" `@{node_label}` decorator can be directly invoked."
                )
            if self.node_def.compute_fn.has_context_arg():
                if len(args) == 0:
                    raise DagsterInvalidInvocationError(
                        f"Compute function of {node_label} '{self.given_alias}' has context"
                        " argument, but no context was provided when invoking."
                    )
                elif args[0] is not None and not isinstance(args[0], UnboundOpExecutionContext):
                    raise DagsterInvalidInvocationError(
                        f"Compute function of {node_label} '{self.given_alias}' has context"
                        " argument, but no context was provided when invoking."
                    )
                context = args[0]
                return op_invocation_result(_self, context, *args[1:], **kwargs)
            else:
                if len(args) > 0 and isinstance(args[0], UnboundOpExecutionContext):
                    raise DagsterInvalidInvocationError(
                        f"Compute function of {node_label} '{self.given_alias}' has no context"
                        " argument, but context was provided when invoking."
                    )
                return op_invocation_result(_self, None, *args, **kwargs)

        assert_in_composition(node_name, self.node_def)
        input_bindings: Dict[str, InputSource] = {}

        # handle *args
        for idx, output_node in enumerate(args):
            if idx >= len(self.node_def.input_defs):
                raise DagsterInvalidDefinitionError(
                    "In {source} {name}, received too many inputs for "
                    "invocation {node_name}. Only {def_num} defined, received {arg_num}".format(
                        source=current_context().source,
                        name=current_context().name,
                        node_name=node_name,
                        def_num=len(self.node_def.input_defs),
                        arg_num=len(args),
                    )
                )

            input_name = self.node_def.resolve_input_name_at_position(idx)
            if input_name is None:
                raise DagsterInvalidDefinitionError(
                    "In {source} {name}, could not resolve input based on position at "
                    "index {idx} for invocation {node_name}. Use keyword args instead, "
                    "available inputs are: {inputs}".format(
                        idx=idx,
                        source=current_context().source,
                        name=current_context().name,
                        node_name=node_name,
                        inputs=list(map(lambda inp: inp.name, self.node_def.input_defs)),
                    )
                )

            self._process_argument_node(
                node_name,
                output_node,
                input_name,
                input_bindings,
                f"(at position {idx})",
            )

        # then **kwargs
        for input_name, output_node in kwargs.items():
            self._process_argument_node(
                node_name,
                output_node,
                input_name,
                input_bindings,
                "(passed by keyword)",
            )

        # the node name is potentially reassigned for aliasing
        resolved_node_name = current_context().observe_invocation(
            self.given_alias,
            self.node_def,
            input_bindings,
            self.tags,
            self.hook_defs,
            self.retry_policy,
        )

        if len(self.node_def.output_defs) == 0:
            return None

        if len(self.node_def.output_defs) == 1:
            output_def = self.node_def.output_defs[0]
            output_name = output_def.name
            if output_def.is_dynamic:
                return InvokedNodeDynamicOutputWrapper(
                    resolved_node_name, output_name, self.node_def.node_type_str
                )
            else:
                return InvokedNodeOutputHandle(
                    resolved_node_name, output_name, self.node_def.node_type_str
                )

        outputs = [output_def for output_def in self.node_def.output_defs]
        invoked_output_handles: Dict[
            str, Union[InvokedNodeDynamicOutputWrapper, InvokedNodeOutputHandle]
        ] = {}
        for output_def in outputs:
            if output_def.is_dynamic:
                invoked_output_handles[output_def.name] = InvokedNodeDynamicOutputWrapper(
                    resolved_node_name, output_def.name, self.node_def.node_type_str
                )
            else:
                invoked_output_handles[output_def.name] = InvokedNodeOutputHandle(
                    resolved_node_name, output_def.name, self.node_def.node_type_str
                )

        return namedtuple(
            f"_{self.node_def.name}_outputs",
            " ".join([output_def.name for output_def in outputs]),
        )(**invoked_output_handles)

    def describe_node(self) -> str:
        node_name = self.given_alias if self.given_alias else self.node_def.name
        return f"{self.node_def.node_type_str} '{node_name}'"

    def _process_argument_node(
        self, node_name: str, output_node, input_name: str, input_bindings, arg_desc: str
    ) -> None:
        from .source_asset import SourceAsset

        # already set - conflict between kwargs and args
        if input_bindings.get(input_name):
            raise DagsterInvalidInvocationError(
                f"{self.node_def.node_type_str} {node_name} got multiple values for"
                f" argument '{input_name}'"
            )

        if isinstance(
            output_node, (InvokedNodeOutputHandle, InputMappingNode, DynamicFanIn, SourceAsset)
        ):
            input_bindings[input_name] = output_node

        elif isinstance(output_node, list):
            input_bindings[input_name] = []
            for idx, fanned_in_node in enumerate(output_node):
                if isinstance(fanned_in_node, (InvokedNodeOutputHandle, InputMappingNode)):
                    input_bindings[input_name].append(fanned_in_node)
                else:
                    raise DagsterInvalidDefinitionError(
                        "In {source} {name}, received a list containing an invalid type "
                        'at index {idx} for input "{input_name}" {arg_desc} in '
                        "{node_type} invocation {node_name}. Lists can only contain the "
                        "output from previous op invocations or input mappings, "
                        "received {type}".format(
                            source=current_context().source,
                            name=current_context().name,
                            arg_desc=arg_desc,
                            input_name=input_name,
                            node_type=self.node_def.node_type_str,
                            node_name=node_name,
                            idx=idx,
                            type=type(output_node),
                        )
                    )

        elif is_named_tuple_instance(output_node) and all(
            map(lambda item: isinstance(item, InvokedNodeOutputHandle), output_node)
        ):
            raise DagsterInvalidDefinitionError(
                "In {source} {name}, received a tuple of multiple outputs for "
                'input "{input_name}" {arg_desc} in {node_type} invocation {node_name}. '
                "Must pass individual output, available from tuple: {options}".format(
                    source=current_context().source,
                    name=current_context().name,
                    arg_desc=arg_desc,
                    input_name=input_name,
                    node_name=node_name,
                    node_type=self.node_def.node_type_str,
                    options=output_node._fields,
                )
            )
        elif isinstance(output_node, InvokedNodeDynamicOutputWrapper):
            raise DagsterInvalidDefinitionError(
                f"In {current_context().source} {current_context().name}, received the dynamic"
                f" output {output_node.output_name} from {output_node.describe_node()} directly."
                " Dynamic output must be unpacked by invoking map or collect."
            )

        elif isinstance(output_node, (NodeDefinition, PendingNodeInvocation)):
            raise DagsterInvalidDefinitionError(
                "In {source} {name}, received an un-invoked {described_node} "
                " for input "
                '"{input_name}" {arg_desc} in {node_type} invocation "{node_name}". '
                "Did you forget parentheses?".format(
                    source=current_context().source,
                    described_node=output_node.describe_node(),
                    name=current_context().name,
                    arg_desc=arg_desc,
                    input_name=input_name,
                    node_name=node_name,
                    node_type=output_node.describe_node(),
                )
            )
        else:
            raise DagsterInvalidDefinitionError(
                "In {source} {name}, received invalid type {type} for input "
                '"{input_name}" {arg_desc} in {node_type} invocation "{node_name}". '
                "Must pass the output from previous node invocations or inputs to the "
                "composition function as inputs when invoking nodes during composition.".format(
                    source=current_context().source,
                    name=current_context().name,
                    type=type(output_node),
                    arg_desc=arg_desc,
                    input_name=input_name,
                    node_name=node_name,
                    node_type=self.node_def.node_type_str,
                )
            )

    @public
    def alias(self, name: str) -> "PendingNodeInvocation[T_NodeDefinition]":
        return PendingNodeInvocation(
            node_def=self.node_def,
            given_alias=name,
            tags=self.tags,
            hook_defs=self.hook_defs,
            retry_policy=self.retry_policy,
        )

    @public
    def tag(self, tags: Optional[Mapping[str, str]]) -> "PendingNodeInvocation[T_NodeDefinition]":
        tags = validate_tags(tags)
        return PendingNodeInvocation(
            node_def=self.node_def,
            given_alias=self.given_alias,
            tags={**(self.tags or {}), **tags},
            hook_defs=self.hook_defs,
            retry_policy=self.retry_policy,
        )

    @public
    def with_hooks(
        self, hook_defs: AbstractSet[HookDefinition]
    ) -> "PendingNodeInvocation[T_NodeDefinition]":
        hook_defs = check.set_param(hook_defs, "hook_defs", of_type=HookDefinition)
        return PendingNodeInvocation(
            node_def=self.node_def,
            given_alias=self.given_alias,
            tags=self.tags,
            hook_defs=set(hook_defs).union(self.hook_defs),
            retry_policy=self.retry_policy,
        )

    @public
    def with_retry_policy(
        self, retry_policy: RetryPolicy
    ) -> "PendingNodeInvocation[T_NodeDefinition]":
        return PendingNodeInvocation(
            node_def=self.node_def,
            given_alias=self.given_alias,
            tags=self.tags,
            hook_defs=self.hook_defs,
            retry_policy=retry_policy,
        )

    @public
    def to_job(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
        config: Optional[Union[ConfigMapping, Mapping[str, Any], "PartitionedConfig"]] = None,
        tags: Optional[Mapping[str, Any]] = None,
        logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
        executor_def: Optional["ExecutorDefinition"] = None,
        hooks: Optional[AbstractSet[HookDefinition]] = None,
        op_retry_policy: Optional[RetryPolicy] = None,
        version_strategy: Optional[VersionStrategy] = None,
        partitions_def: Optional["PartitionsDefinition"] = None,
        input_values: Optional[Mapping[str, object]] = None,
    ) -> "JobDefinition":
        if not isinstance(self.node_def, GraphDefinition):
            raise DagsterInvalidInvocationError(
                "Attemped to call `to_job` on a non-graph.  Only graphs "
                "constructed using the `@graph` decorator support this method."
            )

        tags = check.opt_mapping_param(tags, "tags", key_type=str)
        hooks = check.opt_set_param(hooks, "hooks", HookDefinition)
        input_values = check.opt_mapping_param(input_values, "input_values")
        op_retry_policy = check.opt_inst_param(op_retry_policy, "op_retry_policy", RetryPolicy)
        job_hooks: Set[HookDefinition] = set()
        job_hooks.update(check.opt_set_param(hooks, "hooks", HookDefinition))
        job_hooks.update(self.hook_defs)
        return self.node_def.to_job(
            name=name or self.given_alias,
            description=description,
            resource_defs=resource_defs,
            config=config,
            tags={**(self.tags or {}), **tags},
            logger_defs=logger_defs,
            executor_def=executor_def,
            hooks=job_hooks,
            op_retry_policy=op_retry_policy,
            version_strategy=version_strategy,
            partitions_def=partitions_def,
            input_values=input_values,
        )

    @public
    def execute_in_process(
        self,
        run_config: Optional[Any] = None,
        instance: Optional["DagsterInstance"] = None,
        resources: Optional[Mapping[str, Any]] = None,
        raise_on_error: bool = True,
        run_id: Optional[str] = None,
        input_values: Optional[Mapping[str, object]] = None,
    ) -> "ExecuteInProcessResult":
        if not isinstance(self.node_def, GraphDefinition):
            raise DagsterInvalidInvocationError(
                "Attemped to call `execute_in_process` on a non-graph.  Only graphs "
                "constructed using the `@graph` decorator support this method."
            )

        from dagster._core.execution.build_resources import wrap_resources_for_execution

        from .executor_definition import execute_in_process_executor
        from .job_definition import JobDefinition

        input_values = check.opt_mapping_param(input_values, "input_values")

        ephemeral_job = JobDefinition(
            name=self.given_alias,
            graph_def=self.node_def,
            executor_def=execute_in_process_executor,
            resource_defs=wrap_resources_for_execution(resources),
            tags=self.tags,
            hook_defs=self.hook_defs,
            op_retry_policy=self.retry_policy,
            input_values=input_values,
        )

        return ephemeral_job.execute_in_process(
            run_config=run_config,
            instance=instance,
            raise_on_error=raise_on_error,
            run_id=run_id,
        )


class InvokedNode(NamedTuple):
    """The metadata about a node invocation saved by the current composition context."""

    node_name: str
    node_def: NodeDefinition
    input_bindings: Mapping[str, InputSource]
    tags: Optional[Mapping[str, str]]
    hook_defs: Optional[AbstractSet[HookDefinition]]
    retry_policy: Optional[RetryPolicy]


class InvokedNodeDynamicOutputWrapper:
    """The return value for a dynamic output when invoking a node in a composition function.
    Must be unwrapped by invoking map or collect.
    """

    def __init__(self, node_name: str, output_name: str, node_type: str):
        self.node_name = check.str_param(node_name, "node_name")
        self.output_name = check.str_param(output_name, "output_name")
        self.node_type = check.str_param(node_type, "node_type")

    def describe_node(self) -> str:
        return f"{self.node_type} '{self.node_name}'"

    def map(
        self, fn: Callable
    ) -> Union[
        "InvokedNodeDynamicOutputWrapper", Tuple["InvokedNodeDynamicOutputWrapper", ...], None
    ]:
        check.is_callable(fn)
        result = fn(InvokedNodeOutputHandle(self.node_name, self.output_name, self.node_type))

        if isinstance(result, InvokedNodeOutputHandle):
            return InvokedNodeDynamicOutputWrapper(
                result.node_name, result.output_name, result.node_type
            )
        elif isinstance(result, tuple) and all(
            map(lambda item: isinstance(item, InvokedNodeOutputHandle), result)
        ):
            return tuple(
                map(
                    lambda item: InvokedNodeDynamicOutputWrapper(
                        item.node_name, item.output_name, item.node_type
                    ),
                    result,
                )
            )
        elif result is None:
            return None
        elif isinstance(result, InvokedNodeDynamicOutputWrapper):
            return result
        else:
            check.failed(
                "Could not handle output from map function invoked on "
                f"{self.node_name}:{self.output_name}, received {result}"
            )

    def collect(self) -> DynamicFanIn:
        return DynamicFanIn(self.node_name, self.output_name)

    def unwrap_for_composite_mapping(self) -> InvokedNodeOutputHandle:
        return InvokedNodeOutputHandle(self.node_name, self.output_name, self.node_type)

    def __iter__(self) -> NoReturn:
        raise DagsterInvariantViolationError(
            'Attempted to iterate over an {cls}. This object represents the dynamic output "{out}" '
            'from the {described_node}. Use the "map" method on this object to create '
            "downstream dependencies that will be cloned for each DynamicOut "
            "that is resolved at runtime.".format(
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
            )
        )

    def __getitem__(self, idx) -> NoReturn:
        raise DagsterInvariantViolationError(
            'Attempted to index in to an {cls}. This object represents the dynamic out "{out}" '
            'from the {described_node}. Use the "map" method on this object to create '
            "downstream dependencies that will be cloned for each DynamicOut "
            "that is resolved at runtime.".format(
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
            )
        )

    def alias(self, _) -> NoReturn:
        raise DagsterInvariantViolationError(
            "In {source} {name}, attempted to call alias method for {cls}. This object represents"
            ' the dynamic out "{out}" from the already invoked {described_node}. Consider checking'
            " the location of parentheses.".format(
                source=current_context().source,
                name=current_context().name,
                cls=self.__class__.__name__,
                described_node=self.describe_node(),
                out=self.output_name,
            )
        )

    def with_hooks(self, _) -> NoReturn:
        raise DagsterInvariantViolationError(
            "In {source} {name}, attempted to call hook method for {cls}. This object represents"
            ' the dynamic out "{out}" from the already invoked {described_node}. Consider checking'
            " the location of parentheses.".format(
                source=current_context().source,
                name=current_context().name,
                cls=self.__class__.__name__,
                out=self.output_name,
                described_node=self.describe_node(),
            )
        )


def composite_mapping_from_output(
    output: Any,
    output_defs: Sequence[OutputDefinition],
    node_name: str,
    decorator_name: str,
) -> Optional[Mapping[str, OutputMapping]]:
    # single output
    if isinstance(output, InvokedNodeOutputHandle):
        if len(output_defs) == 1:
            defn = output_defs[0]
            return {defn.name: defn.mapping_from(output.node_name, output.output_name)}
        else:
            raise DagsterInvalidDefinitionError(
                f"Returned a single output ({output.node_name}.{output.output_name}) in "
                f"{decorator_name} '{node_name}' but {len(output_defs)} outputs are defined. "
                "Return a dict to map defined outputs."
            )

    elif isinstance(output, InvokedNodeDynamicOutputWrapper):
        if len(output_defs) == 1:
            defn = output_defs[0]
            return {
                defn.name: defn.mapping_from(
                    output.node_name, output.output_name, from_dynamic_mapping=True
                )
            }
        else:
            raise DagsterInvalidDefinitionError(
                f"Returned a single output ({output.node_name}.{output.output_name}) in "
                f"{decorator_name} '{node_name}' but {len(output_defs)} outputs are defined. "
                "Return a dict to map defined outputs."
            )

    output_mapping_dict = {}
    output_def_dict = {output_def.name: output_def for output_def in output_defs}

    # tuple returned directly
    if isinstance(output, tuple) and all(
        map(lambda item: isinstance(item, InvokedNodeOutputHandle), output)
    ):
        for i, output_name in enumerate(output_def_dict.keys()):
            handle = output[i]
            # map output defined on graph to the actual output defined on the op
            output_mapping_dict[output_name] = output_def_dict[output_name].mapping_from(
                handle.node_name, handle.output_name
            )

        return output_mapping_dict

    # mapping dict
    if isinstance(output, dict):
        for name, handle in output.items():
            if name not in output_def_dict:
                raise DagsterInvalidDefinitionError(
                    f"{decorator_name} '{node_name}' referenced key {name} which does not match any"
                    f" defined outputs. Valid options are: {list(output_def_dict.keys())}"
                )

            if isinstance(handle, InvokedNodeOutputHandle):
                output_mapping_dict[name] = output_def_dict[name].mapping_from(
                    handle.node_name, handle.output_name
                )
            elif isinstance(handle, InvokedNodeDynamicOutputWrapper):
                output_mapping_dict[name] = output_def_dict[name].mapping_from(
                    handle.node_name, handle.output_name, from_dynamic_mapping=True
                )
            else:
                raise DagsterInvalidDefinitionError(
                    f"{decorator_name} '{node_name}' returned problematic dict entry under "
                    f"key {name} of type {type(handle)}. Dict values must be outputs of "
                    "invoked nodes"
                )

        return output_mapping_dict

    # error
    if output is not None:
        raise DagsterInvalidDefinitionError(
            f"{decorator_name} '{node_name}' returned problematic value "
            f"of type {type(output)}. Expected return value from invoked node or dict mapping "
            "output name to return values from invoked nodes"
        )

    return None


def do_composition(
    decorator_name: str,
    graph_name: str,
    fn: Callable,
    provided_input_defs: Sequence[InputDefinition],
    provided_output_defs: Optional[Sequence[OutputDefinition]],
    config_mapping: Optional[ConfigMapping],
    ignore_output_from_composition_fn: bool,
) -> Tuple[
    Sequence[InputMapping],
    Sequence[OutputMapping],
    DependencyMapping[NodeInvocation],
    Sequence[NodeDefinition],
    Optional[ConfigMapping],
    Sequence[str],
    Mapping[str, Mapping[str, "SourceAsset"]],
]:
    """This a function used by both @job and @graph to implement their composition
    function which is our DSL for constructing a dependency graph.

    Args:
        decorator_name (str): Name of the calling decorator. e.g. "@graph" or "@job"
        graph_name (str): User-defined name of the definition being constructed
        fn (Callable): The composition function to be called.
        provided_input_defs(List[InputDefinition]): List of input definitions
            explicitly provided to the decorator by the user.
        provided_output_defs(List[OutputDefinition]): List of output definitions
            explicitly provided to the decorator by the user.
        config_mapping (Any): Config mapping provided to decorator by user. In
            job/graph case, this would have been constructed from a user-provided
            config_schema and config_fn.
        ignore_output_from_composite_fn(Bool): Because of backwards compatibility
            issues, jobs ignore the return value out of the mapping if
            the user has not explicitly provided the output definitions.
            This should be removed in 0.11.0.
    """
    from .decorators.op_decorator import (
        NoContextDecoratedOpFunction,
        resolve_checked_op_fn_inputs,
    )

    actual_output_defs: Sequence[OutputDefinition]
    if provided_output_defs is None:
        outputs_are_explicit = False
        actual_output_defs = [OutputDefinition.create_from_inferred(infer_output_props(fn))]
    elif len(provided_output_defs) == 1:
        outputs_are_explicit = True
        actual_output_defs = [provided_output_defs[0].combine_with_inferred(infer_output_props(fn))]
    else:
        outputs_are_explicit = True
        actual_output_defs = provided_output_defs

    compute_fn = NoContextDecoratedOpFunction(fn)

    actual_input_defs = resolve_checked_op_fn_inputs(
        decorator_name=decorator_name,
        fn_name=graph_name,
        compute_fn=compute_fn,
        explicit_input_defs=provided_input_defs,
        exclude_nothing=False,
    )

    kwargs = {input_def.name: InputMappingNode(input_def) for input_def in actual_input_defs}

    output = None
    returned_mapping = None
    enter_composition(graph_name, decorator_name)
    try:
        output = fn(**kwargs)
        if ignore_output_from_composition_fn:
            output = None

        returned_mapping = composite_mapping_from_output(
            output, actual_output_defs, graph_name, decorator_name
        )
    finally:
        context = exit_composition(returned_mapping)

    check.invariant(
        context.name == graph_name,
        "Composition context stack desync: received context for "
        '"{context.name}" expected "{graph_name}"'.format(context=context, graph_name=graph_name),
    )

    # line up mappings in definition order
    input_mappings = []
    for defn in actual_input_defs:
        mappings = [
            mapping for mapping in context.input_mappings if mapping.graph_input_name == defn.name
        ]

        if len(mappings) == 0:
            raise DagsterInvalidDefinitionError(
                "{decorator_name} '{graph_name}' has unmapped input '{input_name}'. "
                "Remove it or pass it to the appropriate op/graph invocation.".format(
                    decorator_name=decorator_name,
                    graph_name=graph_name,
                    input_name=defn.name,
                )
            )

        input_mappings += mappings

    output_mappings = []
    for defn in actual_output_defs:
        mapping = context.output_mapping_dict.get(defn.name)
        if mapping is None:
            # if we inferred output_defs we will be flexible and either take a mapping or not
            if not outputs_are_explicit:
                continue

            # if we are ignoring the output, disregard this unsatisfied mapping
            if ignore_output_from_composition_fn:
                continue

            raise DagsterInvalidDefinitionError(
                "{decorator_name} '{graph_name}' has unmapped output '{output_name}'. "
                "Remove it or return a value from the appropriate op/graph invocation.".format(
                    decorator_name=decorator_name, graph_name=graph_name, output_name=defn.name
                )
            )
        output_mappings.append(mapping)

    return (
        input_mappings,
        output_mappings,
        context.dependencies,
        context.node_defs,
        config_mapping,
        compute_fn.positional_inputs(),
        context.node_input_source_assets,
    )


def get_validated_config_mapping(
    name: str,
    config_schema: Any,
    config_fn: Optional[Callable[[Any], Any]],
    decorator_name: str,
) -> Optional[ConfigMapping]:
    if config_fn is None and config_schema is None:
        return None
    elif config_fn is not None:
        return ConfigMapping(config_fn=config_fn, config_schema=config_schema)
    else:
        raise DagsterInvalidDefinitionError(
            f"{decorator_name} '{name}' defines a configuration schema but does not "
            "define a configuration function."
        )



from .base import ScheduleStorage as ScheduleStorage
from .schema import ScheduleStorageSqlMetadata as ScheduleStorageSqlMetadata
from .sql_schedule_storage import SqlScheduleStorage as SqlScheduleStorage
from .sqlite import SqliteScheduleStorage as SqliteScheduleStorage



from typing import TYPE_CHECKING, Callable, Dict, Mapping, NamedTuple, Optional, Set, cast

import pendulum

import dagster._check as check
from dagster._annotations import PublicAttr, experimental
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.scoped_resources_builder import Resources, ScopedResourcesBuilder
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    FreshnessPolicySensorExecutionError,
    user_code_error_boundary,
)
from dagster._core.instance import DagsterInstance
from dagster._serdes import (
    serialize_value,
    whitelist_for_serdes,
)
from dagster._serdes.errors import DeserializationError
from dagster._serdes.serdes import deserialize_value
from dagster._seven import JSONDecodeError

from .sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorEvaluationContext,
    SensorType,
    SkipReason,
    get_context_param_name,
    get_sensor_context_from_args_or_kwargs,
    validate_and_get_resource_dict,
)

if TYPE_CHECKING:
    pass


@whitelist_for_serdes
class FreshnessPolicySensorCursor(
    NamedTuple(
        "_FreshnessPolicySensorCursor",
        [("minutes_late_by_key_str", Mapping[str, Optional[float]])],
    )
):
    def __new__(cls, minutes_late_by_key_str: Mapping[str, Optional[float]]):
        return super(FreshnessPolicySensorCursor, cls).__new__(
            cls,
            minutes_late_by_key_str=check.mapping_param(
                minutes_late_by_key_str, "minutes_late_by_key_str", key_type=str
            ),
        )

    @staticmethod
    def is_valid(json_str: str) -> bool:
        try:
            deserialize_value(json_str, FreshnessPolicySensorCursor)
            return True
        except (JSONDecodeError, DeserializationError):
            return False

    @staticmethod
    def from_dict(
        minutes_late_by_key: Mapping[AssetKey, Optional[float]]
    ) -> "FreshnessPolicySensorCursor":
        return FreshnessPolicySensorCursor(
            minutes_late_by_key_str={k.to_user_string(): v for k, v in minutes_late_by_key.items()}
        )

    @property
    def minutes_late_by_key(self) -> Mapping[AssetKey, Optional[float]]:
        return {AssetKey.from_user_string(k): v for k, v in self.minutes_late_by_key_str.items()}

    def to_json(self) -> str:
        return serialize_value(cast(NamedTuple, self))

    @staticmethod
    def from_json(json_str: str) -> "FreshnessPolicySensorCursor":
        return deserialize_value(json_str, FreshnessPolicySensorCursor)


class FreshnessPolicySensorContext(
    NamedTuple(
        "_FreshnessPolicySensorContext",
        [
            ("sensor_name", PublicAttr[str]),
            ("asset_key", PublicAttr[AssetKey]),
            ("freshness_policy", PublicAttr[FreshnessPolicy]),
            ("minutes_overdue", PublicAttr[Optional[float]]),
            ("previous_minutes_overdue", PublicAttr[Optional[float]]),
            ("instance", PublicAttr[DagsterInstance]),
            ("resources", Resources),
        ],
    )
):
    """The ``context`` object available to a decorated function of ``freshness_policy_sensor``.

    Attributes:
        sensor_name (str): the name of the sensor.
        asset_key (AssetKey): the key of the asset being monitored
        freshness_policy (FreshnessPolicy): the freshness policy of the asset being monitored
        minutes_overdue (Optional[float])
        previous_minutes_overdue (Optional[float]): the minutes_overdue value for this asset on the
            previous sensor tick.
        instance (DagsterInstance): the current instance.
    """

    def __new__(
        cls,
        sensor_name: str,
        asset_key: AssetKey,
        freshness_policy: FreshnessPolicy,
        minutes_overdue: Optional[float],
        previous_minutes_overdue: Optional[float],
        instance: DagsterInstance,
        resources: Optional[Resources] = None,
    ):
        minutes_overdue = check.opt_numeric_param(minutes_overdue, "minutes_overdue")
        previous_minutes_overdue = check.opt_numeric_param(
            previous_minutes_overdue, "previous_minutes_overdue"
        )
        return super(FreshnessPolicySensorContext, cls).__new__(
            cls,
            sensor_name=check.str_param(sensor_name, "sensor_name"),
            asset_key=check.inst_param(asset_key, "asset_key", AssetKey),
            freshness_policy=check.inst_param(freshness_policy, "FreshnessPolicy", FreshnessPolicy),
            minutes_overdue=float(minutes_overdue) if minutes_overdue is not None else None,
            previous_minutes_overdue=float(previous_minutes_overdue)
            if previous_minutes_overdue is not None
            else None,
            instance=check.inst_param(instance, "instance", DagsterInstance),
            resources=resources or ScopedResourcesBuilder.build_empty(),
        )


@experimental
def build_freshness_policy_sensor_context(
    sensor_name: str,
    asset_key: AssetKey,
    freshness_policy: FreshnessPolicy,
    minutes_overdue: Optional[float],
    previous_minutes_overdue: Optional[float] = None,
    instance: Optional[DagsterInstance] = None,
    resources: Optional[Resources] = None,
) -> FreshnessPolicySensorContext:
    """Builds freshness policy sensor context from provided parameters.

    This function can be used to provide the context argument when directly invoking a function
    decorated with `@freshness_policy_sensor`, such as when writing unit tests.

    Args:
        sensor_name (str): The name of the sensor the context is being constructed for.
        asset_key (AssetKey): The AssetKey for the monitored asset
        freshness_policy (FreshnessPolicy): The FreshnessPolicy for the monitored asset
        minutes_overdue (Optional[float]): How overdue the monitored asset currently is
        previous_minutes_overdue (Optional[float]): How overdue the monitored asset was on the
            previous tick.
        instance (DagsterInstance): The dagster instance configured for the context.

    Examples:
        .. code-block:: python

            context = build_freshness_policy_sensor_context(
                sensor_name="freshness_policy_sensor_to_invoke",
                asset_key=AssetKey("some_asset"),
                freshness_policy=FreshnessPolicy(maximum_lag_minutes=30)<
                minutes_overdue=10.0,
            )
            freshness_policy_sensor_to_invoke(context)
    """
    return FreshnessPolicySensorContext(
        sensor_name=sensor_name,
        asset_key=asset_key,
        freshness_policy=freshness_policy,
        minutes_overdue=minutes_overdue,
        previous_minutes_overdue=previous_minutes_overdue,
        instance=instance or DagsterInstance.ephemeral(),
        resources=resources,
    )


class FreshnessPolicySensorDefinition(SensorDefinition):
    """Define a sensor that reacts to the status of a given set of asset freshness policies,
    where the decorated function will be evaluated on every sensor tick.

    Args:
        name (str): The name of the sensor. Defaults to the name of the decorated function.
        freshness_policy_sensor_fn (Callable[[FreshnessPolicySensorContext], None]): The core
            evaluation function for the sensor. Takes a :py:class:`~dagster.FreshnessPolicySensorContext`.
        asset_selection (AssetSelection): The asset selection monitored by the sensor.
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from Dagit or via the GraphQL API.
    """

    def __init__(
        self,
        name: str,
        asset_selection: AssetSelection,
        freshness_policy_sensor_fn: Callable[..., None],
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        required_resource_keys: Optional[Set[str]] = None,
    ):
        check.str_param(name, "name")
        check.inst_param(asset_selection, "asset_selection", AssetSelection)
        check.opt_int_param(minimum_interval_seconds, "minimum_interval_seconds")
        check.opt_str_param(description, "description")
        check.inst_param(default_status, "default_status", DefaultSensorStatus)

        self._freshness_policy_sensor_fn = check.callable_param(
            freshness_policy_sensor_fn, "freshness_policy_sensor_fn"
        )

        resource_arg_names: Set[str] = {
            arg.name for arg in get_resource_args(freshness_policy_sensor_fn)
        }

        combined_required_resource_keys = (
            check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
            | resource_arg_names
        )

        def _wrapped_fn(context: SensorEvaluationContext):
            from dagster._utils.caching_instance_queryer import (
                CachingInstanceQueryer,  # expensive import
            )

            if context.repository_def is None:
                raise DagsterInvalidInvocationError(
                    "The `repository_def` property on the `SensorEvaluationContext` passed into a "
                    "`FreshnessPolicySensorDefinition` must not be None."
                )

            if context.cursor is None or not FreshnessPolicySensorCursor.is_valid(context.cursor):
                new_cursor = FreshnessPolicySensorCursor({})
                context.update_cursor(new_cursor.to_json())
                yield SkipReason(f"Initializing {name}.")
                return

            evaluation_time = pendulum.now("UTC")
            asset_graph = context.repository_def.asset_graph
            instance_queryer = CachingInstanceQueryer(
                context.instance, asset_graph, evaluation_time
            )
            data_time_resolver = CachingDataTimeResolver(instance_queryer=instance_queryer)
            monitored_keys = asset_selection.resolve(asset_graph)

            # get the previous status from the cursor
            previous_minutes_late_by_key = FreshnessPolicySensorCursor.from_json(
                context.cursor
            ).minutes_late_by_key

            minutes_late_by_key: Dict[AssetKey, Optional[float]] = {}
            for asset_key in monitored_keys:
                freshness_policy = asset_graph.freshness_policies_by_key.get(asset_key)
                if freshness_policy is None:
                    continue

                # get the current minutes_overdue value for this asset
                minutes_late_by_key[asset_key] = data_time_resolver.get_current_minutes_late(
                    evaluation_time=evaluation_time,
                    asset_key=asset_key,
                )

                resource_args_populated = validate_and_get_resource_dict(
                    context.resources, name, resource_arg_names
                )
                context_param_name = get_context_param_name(freshness_policy_sensor_fn)
                freshness_context = FreshnessPolicySensorContext(
                    sensor_name=name,
                    asset_key=asset_key,
                    freshness_policy=freshness_policy,
                    minutes_overdue=minutes_late_by_key[asset_key],
                    previous_minutes_overdue=previous_minutes_late_by_key.get(asset_key),
                    instance=context.instance,
                    resources=context.resources,
                )

                with user_code_error_boundary(
                    FreshnessPolicySensorExecutionError,
                    lambda: f'Error occurred during the execution of sensor "{name}".',
                ):
                    context_param = (
                        {context_param_name: freshness_context} if context_param_name else {}
                    )
                    result = freshness_policy_sensor_fn(
                        **context_param,
                        **resource_args_populated,
                    )
                if result is not None:
                    raise DagsterInvalidDefinitionError(
                        "Functions decorated by `@freshness_policy_sensor` may not return or yield"
                        " a value."
                    )

            context.update_cursor(
                FreshnessPolicySensorCursor.from_dict(minutes_late_by_key).to_json()
            )

        super(FreshnessPolicySensorDefinition, self).__init__(
            name=name,
            evaluation_fn=_wrapped_fn,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            default_status=default_status,
            required_resource_keys=combined_required_resource_keys,
        )

    def __call__(self, *args, **kwargs) -> None:
        context_param_name = get_context_param_name(self._freshness_policy_sensor_fn)

        sensor_context = get_sensor_context_from_args_or_kwargs(
            self._freshness_policy_sensor_fn,
            args,
            kwargs,
            context_type=FreshnessPolicySensorContext,
        )
        context_param = (
            {context_param_name: sensor_context} if context_param_name and sensor_context else {}
        )

        resources = validate_and_get_resource_dict(
            sensor_context.resources if sensor_context else ScopedResourcesBuilder.build_empty(),
            self._name,
            self._required_resource_keys,
        )

        return self._freshness_policy_sensor_fn(**context_param, **resources)

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.FRESHNESS_POLICY


@experimental
def freshness_policy_sensor(
    asset_selection: AssetSelection,
    *,
    name: Optional[str] = None,
    minimum_interval_seconds: Optional[int] = None,
    description: Optional[str] = None,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
) -> Callable[[Callable[..., None]], FreshnessPolicySensorDefinition,]:
    """Define a sensor that reacts to the status of a given set of asset freshness policies, where the
    decorated function will be evaluated on every tick for each asset in the selection that has a
    FreshnessPolicy defined.

    Note: returning or yielding a value from the annotated function will result in an error.

    Takes a :py:class:`~dagster.FreshnessPolicySensorContext`.

    Args:
        asset_selection (AssetSelection): The asset selection monitored by the sensor.
        name (Optional[str]): The name of the sensor. Defaults to the name of the decorated function.
        freshness_policy_sensor_fn (Callable[[FreshnessPolicySensorContext], None]): The core
            evaluation function for the sensor. Takes a :py:class:`~dagster.FreshnessPolicySensorContext`.
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from Dagit or via the GraphQL API.
    """

    def inner(fn: Callable[..., None]) -> FreshnessPolicySensorDefinition:
        check.callable_param(fn, "fn")
        sensor_name = name or fn.__name__

        return FreshnessPolicySensorDefinition(
            name=sensor_name,
            freshness_policy_sensor_fn=fn,
            asset_selection=asset_selection,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            default_status=default_status,
        )

    return inner



from .step_delegating_executor import StepDelegatingExecutor as StepDelegatingExecutor
from .step_handler import (
    CheckStepHealthResult as CheckStepHealthResult,
    StepHandler as StepHandler,
    StepHandlerContext as StepHandlerContext,
)



import json
from collections import defaultdict
from inspect import isfunction
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Union,
    cast,
)

import dagster._check as check
from dagster._config.pythonic_config import (
    ConfigurableIOManagerFactoryResourceDefinition,
    ConfigurableResourceFactoryResourceDefinition,
    ResourceWithKeyMapping,
    coerce_to_resource,
)
from dagster._core.definitions.asset_graph import AssetGraph
from dagster._core.definitions.assets_job import (
    get_base_asset_jobs,
    is_base_asset_job_name,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.executor_definition import ExecutorDefinition
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.partitioned_schedule import (
    UnresolvedPartitionedAssetScheduleDefinition,
)
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.errors import DagsterInvalidDefinitionError

from .repository_data import CachingRepositoryData
from .valid_definitions import VALID_REPOSITORY_DATA_DICT_KEYS, RepositoryListDefinition


def _find_env_vars(config_entry: Any) -> Set[str]:
    """Given a part of a config dictionary, return a set of environment variables that are used in
    that part of the config.
    """
    # Actual env var entry
    if isinstance(config_entry, Mapping) and set(config_entry.keys()) == {"env"}:
        return {config_entry["env"]}
    # Recurse into dictionary of config items
    elif isinstance(config_entry, Mapping):
        return set().union(*[_find_env_vars(v) for v in config_entry.values()])
    # Recurse into list of config items
    elif isinstance(config_entry, List):
        return set().union(*[_find_env_vars(v) for v in config_entry])

    # Otherwise, raw config value which is not an env var, so return empty set
    return set()


def _env_vars_from_resource_defaults(resource_def: ResourceDefinition) -> Set[str]:
    """Given a resource definition, return a set of environment variables that are used in the
    resource's default config. This is used to extract environment variables from the top-level
    resources in a Definitions object.
    """
    config_schema_default = cast(
        Mapping[str, Any],
        json.loads(resource_def.config_schema.default_value_as_json_str)
        if resource_def.config_schema.default_provided
        else {},
    )

    env_vars = _find_env_vars(config_schema_default)

    if isinstance(resource_def, ResourceWithKeyMapping) and isinstance(
        resource_def.inner_resource,
        (
            ConfigurableIOManagerFactoryResourceDefinition,
            ConfigurableResourceFactoryResourceDefinition,
        ),
    ):
        nested_resources = resource_def.inner_resource.nested_resources
        for nested_resource in nested_resources.values():
            env_vars = env_vars.union(
                _env_vars_from_resource_defaults(coerce_to_resource(nested_resource))
            )

    return env_vars


def build_caching_repository_data_from_list(
    repository_definitions: Sequence[RepositoryListDefinition],
    default_executor_def: Optional[ExecutorDefinition] = None,
    default_logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
    top_level_resources: Optional[Mapping[str, ResourceDefinition]] = None,
    resource_key_mapping: Optional[Mapping[int, str]] = None,
) -> CachingRepositoryData:
    from dagster._core.definitions import AssetsDefinition
    from dagster._core.definitions.partitioned_schedule import (
        UnresolvedPartitionedAssetScheduleDefinition,
    )

    schedule_and_sensor_names: Set[str] = set()
    jobs: Dict[str, JobDefinition] = {}
    coerced_graphs: Dict[str, JobDefinition] = {}
    unresolved_jobs: Dict[str, UnresolvedAssetJobDefinition] = {}
    schedules: Dict[str, ScheduleDefinition] = {}
    unresolved_partitioned_asset_schedules: Dict[
        str, UnresolvedPartitionedAssetScheduleDefinition
    ] = {}
    sensors: Dict[str, SensorDefinition] = {}
    assets_defs: List[AssetsDefinition] = []
    asset_keys: Set[AssetKey] = set()
    source_assets: List[SourceAsset] = []
    for definition in repository_definitions:
        if isinstance(definition, JobDefinition):
            if (
                definition.name in jobs and jobs[definition.name] != definition
            ) or definition.name in unresolved_jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate job definition found for {definition.describe_target()}"
                )
            if is_base_asset_job_name(definition.name):
                raise DagsterInvalidDefinitionError(
                    f"Attempted to provide job called {definition.name} to repository, which "
                    "is a reserved name. Please rename the job."
                )
            jobs[definition.name] = definition
        elif isinstance(definition, SensorDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)
            sensors[definition.name] = definition
        elif isinstance(definition, ScheduleDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)

            schedules[definition.name] = definition

        elif isinstance(definition, UnresolvedPartitionedAssetScheduleDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)

            unresolved_partitioned_asset_schedules[definition.name] = definition
        elif isinstance(definition, GraphDefinition):
            coerced = definition.coerce_to_job()
            if coerced.name in jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate job definition found for graph '{coerced.name}'"
                )
            jobs[coerced.name] = coerced
            coerced_graphs[coerced.name] = coerced
        elif isinstance(definition, UnresolvedAssetJobDefinition):
            if definition.name in jobs or definition.name in unresolved_jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for unresolved job '{definition.name}'"
                )
            # we can only resolve these once we have all assets
            unresolved_jobs[definition.name] = definition
        elif isinstance(definition, AssetsDefinition):
            for key in definition.keys:
                if key in asset_keys:
                    raise DagsterInvalidDefinitionError(f"Duplicate asset key: {key}")

            asset_keys.update(definition.keys)
            assets_defs.append(definition)
        elif isinstance(definition, SourceAsset):
            source_assets.append(definition)
        else:
            check.failed(f"Unexpected repository entry {definition}")

    if assets_defs or source_assets:
        for job_def in get_base_asset_jobs(
            assets=assets_defs,
            source_assets=source_assets,
            executor_def=default_executor_def,
            resource_defs=top_level_resources,
        ):
            jobs[job_def.name] = job_def

        source_assets_by_key = {source_asset.key: source_asset for source_asset in source_assets}
        assets_defs_by_key = {key: asset for asset in assets_defs for key in asset.keys}
    else:
        source_assets_by_key = {}
        assets_defs_by_key = {}

    for name, sensor_def in sensors.items():
        if sensor_def.has_loadable_targets():
            targets = sensor_def.load_targets()
            for target in targets:
                _process_and_validate_target(
                    sensor_def, coerced_graphs, unresolved_jobs, jobs, target
                )

    for name, schedule_def in schedules.items():
        if schedule_def.has_loadable_target():
            target = schedule_def.load_target()
            _process_and_validate_target(
                schedule_def, coerced_graphs, unresolved_jobs, jobs, target
            )

    asset_graph = AssetGraph.from_assets([*assets_defs, *source_assets])

    if unresolved_partitioned_asset_schedules:
        for (
            name,
            unresolved_partitioned_asset_schedule,
        ) in unresolved_partitioned_asset_schedules.items():
            _process_and_validate_target(
                unresolved_partitioned_asset_schedule,
                coerced_graphs,
                unresolved_jobs,
                jobs,
                unresolved_partitioned_asset_schedule.job,
            )

    # resolve all the UnresolvedAssetJobDefinitions using the full set of assets
    if unresolved_jobs:
        for name, unresolved_job_def in unresolved_jobs.items():
            resolved_job = unresolved_job_def.resolve(
                asset_graph=asset_graph,
                default_executor_def=default_executor_def,
                resource_defs=top_level_resources,
            )
            jobs[name] = resolved_job

    # resolve all the UnresolvedPartitionedAssetScheduleDefinitions using
    # the resolved job containing the partitions def
    if unresolved_partitioned_asset_schedules:
        for (
            name,
            unresolved_partitioned_asset_schedule,
        ) in unresolved_partitioned_asset_schedules.items():
            resolved_job = jobs[unresolved_partitioned_asset_schedule.job.name]
            schedules[name] = unresolved_partitioned_asset_schedule.resolve(
                cast(JobDefinition, resolved_job)
            )

    for job_def in jobs.values():
        job_def.validate_resource_requirements_satisfied()

    if default_executor_def:
        for name, job_def in jobs.items():
            if not job_def.has_specified_executor:
                jobs[name] = job_def.with_executor_def(default_executor_def)

    if default_logger_defs:
        for name, job_def in jobs.items():
            if not job_def.has_specified_loggers:
                jobs[name] = job_def.with_logger_defs(default_logger_defs)

    top_level_resources = top_level_resources or {}

    utilized_env_vars: Dict[str, Set[str]] = defaultdict(set)

    for resource_key, resource_def in top_level_resources.items():
        used_env_vars = _env_vars_from_resource_defaults(resource_def)
        for env_var in used_env_vars:
            utilized_env_vars[env_var].add(resource_key)

    return CachingRepositoryData(
        jobs=jobs,
        schedules=schedules,
        sensors=sensors,
        source_assets_by_key=source_assets_by_key,
        assets_defs_by_key=assets_defs_by_key,
        top_level_resources=top_level_resources or {},
        utilized_env_vars=utilized_env_vars,
        resource_key_mapping=resource_key_mapping or {},
    )


def build_caching_repository_data_from_dict(
    repository_definitions: Dict[str, Dict[str, Any]]
) -> "CachingRepositoryData":
    check.dict_param(repository_definitions, "repository_definitions", key_type=str)
    check.invariant(
        set(repository_definitions.keys()).issubset(VALID_REPOSITORY_DATA_DICT_KEYS),
        "Bad dict: must not contain keys other than {{{valid_keys}}}: found {bad_keys}.".format(
            valid_keys=", ".join([f"'{key}'" for key in VALID_REPOSITORY_DATA_DICT_KEYS]),
            bad_keys=", ".join(
                [
                    "'{key}'"
                    for key in repository_definitions.keys()
                    if key not in VALID_REPOSITORY_DATA_DICT_KEYS
                ]
            ),
        ),
    )

    for key in VALID_REPOSITORY_DATA_DICT_KEYS:
        if key not in repository_definitions:
            repository_definitions[key] = {}

    duplicate_keys = set(repository_definitions["schedules"].keys()).intersection(
        set(repository_definitions["sensors"].keys())
    )
    if duplicate_keys:
        raise DagsterInvalidDefinitionError(
            "Duplicate definitions between schedules and sensors found for keys:"
            f" {', '.join(duplicate_keys)}"
        )

    for key, raw_job_def in repository_definitions["jobs"].items():
        if isinstance(raw_job_def, GraphDefinition):
            repository_definitions["jobs"][key] = raw_job_def.coerce_to_job()
        elif isinstance(raw_job_def, UnresolvedAssetJobDefinition):
            repository_definitions["jobs"][key] = raw_job_def.resolve(
                # TODO: https://github.com/dagster-io/dagster/issues/8263
                assets=[],
                source_assets=[],
                default_executor_def=None,
            )
        elif not isinstance(raw_job_def, JobDefinition) and not isfunction(raw_job_def):
            raise DagsterInvalidDefinitionError(
                f"Object mapped to {key} is not an instance of JobDefinition or GraphDefinition."
            )

    # Late validate all jobs' resource requirements are satisfied, since
    # they may not be applied until now
    for job_def in repository_definitions["jobs"].values():
        if isinstance(job_def, JobDefinition):
            job_def.validate_resource_requirements_satisfied()

    return CachingRepositoryData(
        **repository_definitions,
        source_assets_by_key={},
        assets_defs_by_key={},
        top_level_resources={},
        utilized_env_vars={},
        resource_key_mapping={},
    )


def _process_and_validate_target(
    schedule_or_sensor_def: Union[
        SensorDefinition, ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition
    ],
    coerced_graphs: Dict[str, JobDefinition],
    unresolved_jobs: Dict[str, UnresolvedAssetJobDefinition],
    jobs: Dict[str, JobDefinition],
    target: Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition],
):
    """This function modifies the state of coerced_graphs, unresolved_jobs, and jobs."""
    targeter = (
        f"schedule '{schedule_or_sensor_def.name}'"
        if isinstance(schedule_or_sensor_def, ScheduleDefinition)
        else f"sensor '{schedule_or_sensor_def.name}'"
    )
    if isinstance(target, GraphDefinition):
        if target.name not in coerced_graphs:
            # Since this is a graph we have to coerce, it is not possible to be
            # the same definition by reference equality
            if target.name in jobs:
                raise DagsterInvalidDefinitionError(
                    _get_error_msg_for_target_conflict(targeter, "graph", target.name, "job")
                )
        elif coerced_graphs[target.name].graph != target:
            raise DagsterInvalidDefinitionError(
                _get_error_msg_for_target_conflict(targeter, "graph", target.name, "graph")
            )
        coerced_job = target.coerce_to_job()
        coerced_graphs[target.name] = coerced_job
        jobs[target.name] = coerced_job
    elif isinstance(target, UnresolvedAssetJobDefinition):
        if target.name not in unresolved_jobs:
            # Since this is an unresolved job we have to resolve, it is not possible to
            # be the same definition by reference equality
            if target.name in jobs:
                raise DagsterInvalidDefinitionError(
                    _get_error_msg_for_target_conflict(
                        targeter, "unresolved asset job", target.name, "job"
                    )
                )
        elif unresolved_jobs[target.name].selection != target.selection:
            raise DagsterInvalidDefinitionError(
                _get_error_msg_for_target_conflict(
                    targeter, "unresolved asset job", target.name, "unresolved asset job"
                )
            )
        unresolved_jobs[target.name] = target
    else:
        if target.name in jobs and jobs[target.name] != target:
            dupe_target_type = (
                "graph"
                if target.name in coerced_graphs
                else "unresolved asset job"
                if target.name in unresolved_jobs
                else "job"
            )
            raise DagsterInvalidDefinitionError(
                _get_error_msg_for_target_conflict(targeter, "job", target.name, dupe_target_type)
            )
        jobs[target.name] = target


def _get_error_msg_for_target_conflict(targeter, target_type, target_name, dupe_target_type):
    return (
        f"{targeter} targets {target_type} '{target_name}', but a different {dupe_target_type} with"
        " the same name was provided. Disambiguate between these by providing a separate name to"
        " one of them."
    )



import threading
from functools import lru_cache
from typing import Any, Optional, Tuple, Union

import sqlalchemy as db
from alembic.command import downgrade, stamp, upgrade
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy.engine import Connection
from sqlalchemy.ext.compiler import compiles
from typing_extensions import TypeAlias

from dagster._utils import file_relative_path

create_engine = db.create_engine  # exported


ALEMBIC_SCRIPTS_LOCATION = "dagster:_core/storage/alembic"

# Stand-in for a typed query object, which is only available in sqlalchemy 2+
SqlAlchemyQuery: TypeAlias = Any

# Stand-in for a typed row object, which is only available in sqlalchemy 2+
SqlAlchemyRow: TypeAlias = Any

AlembicVersion: TypeAlias = Tuple[Optional[str], Optional[Union[str, Tuple[str, ...]]]]


@lru_cache(maxsize=3)  # run, event, and schedule storages
def get_alembic_config(
    dunder_file: str,
    config_path: str = "alembic/alembic.ini",
    script_location: Optional[str] = None,
) -> Config:
    if not script_location:
        script_location = ALEMBIC_SCRIPTS_LOCATION

    alembic_config = Config(file_relative_path(dunder_file, config_path))
    alembic_config.set_main_option("script_location", script_location)
    return alembic_config


def run_alembic_upgrade(
    alembic_config: Config, conn: Connection, run_id: Optional[str] = None, rev: str = "head"
) -> None:
    alembic_config.attributes["connection"] = conn
    alembic_config.attributes["run_id"] = run_id
    upgrade(alembic_config, rev)


def run_alembic_downgrade(
    alembic_config: Config, conn: Connection, rev: str, run_id: Optional[str] = None
) -> None:
    alembic_config.attributes["connection"] = conn
    alembic_config.attributes["run_id"] = run_id
    downgrade(alembic_config, rev)


# Ensure that at most one thread can be stamping alembic revisions at once
_alembic_lock = threading.Lock()


def stamp_alembic_rev(alembic_config: Config, conn: Connection, rev: str = "head") -> None:
    with _alembic_lock:
        alembic_config.attributes["connection"] = conn
        stamp(alembic_config, rev)


def check_alembic_revision(alembic_config: Config, conn: Connection) -> AlembicVersion:
    with _alembic_lock:
        migration_context = MigrationContext.configure(conn)
        db_revision = migration_context.get_current_revision()
        script = ScriptDirectory.from_config(alembic_config)
        head_revision = script.as_revision_number("head")

    return (db_revision, head_revision)


def run_migrations_offline(
    context: EnvironmentContext, config: Config, target_metadata: db.MetaData
) -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    from sqlite3 import DatabaseError

    connectable = config.attributes.get("connection", None)

    if connectable is None:
        raise Exception(
            "No connection set in alembic config. If you are trying to run this script from the "
            "command line, STOP and read the README."
        )

    try:
        context.configure(
            url=connectable.url,
            target_metadata=target_metadata,
            literal_binds=True,
            dialect_opts={"paramstyle": "named"},
        )

        with context.begin_transaction():
            context.run_migrations()
    except DatabaseError as exc:
        # This is to deal with concurrent execution -- if this table already exists thanks to a
        # race with another process, we are fine and can continue.
        if "table alembic_version already exists" not in str(exc):
            raise


def run_migrations_online(
    context: EnvironmentContext, config: Config, target_metadata: db.MetaData
) -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    from sqlite3 import DatabaseError

    connection = config.attributes.get("connection", None)

    if connection is None:
        raise Exception(
            "No connection set in alembic config. If you are trying to run this script from the "
            "command line, STOP and read the README."
        )

    try:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()

    except DatabaseError as exc:
        # This is to deal with concurrent execution -- if this table already exists thanks to a
        # race with another process, we are fine and can continue.
        if "table alembic_version already exists" not in str(exc):
            raise


# SQLAlchemy types, compiler directives, etc. to avoid pre-0.11.0 migrations
# as well as compiler directives to make cross-DB API semantics the same.

# 1: make MySQL dates equivalent to PG or Sqlite dates

MYSQL_DATE_PRECISION: int = 6
MYSQL_FLOAT_PRECISION: int = 32


# datetime issue fix from here: https://stackoverflow.com/questions/29711102/sqlalchemy-mysql-millisecond-or-microsecond-precision/29723278
@compiles(db.DateTime, "mysql")
def compile_datetime_and_add_precision_mysql(_element, _compiler, **_kw) -> str:
    return f"DATETIME({MYSQL_DATE_PRECISION})"


class get_current_timestamp(db.sql.expression.FunctionElement):
    """Like CURRENT_TIMESTAMP, but has the same semantics on MySQL, Postgres, and Sqlite."""

    type = db.types.DateTime()  # type: ignore


@compiles(get_current_timestamp, "mysql")
def compiles_get_current_timestamp_mysql(_element, _compiler, **_kw) -> str:
    return f"CURRENT_TIMESTAMP({MYSQL_DATE_PRECISION})"


@compiles(get_current_timestamp)
def compiles_get_current_timestamp_default(_element, _compiler, **_kw) -> str:
    return "CURRENT_TIMESTAMP"


@compiles(db.types.TIMESTAMP, "mysql")
def add_precision_to_mysql_timestamps(_element, _compiler, **_kw) -> str:
    return f"TIMESTAMP({MYSQL_DATE_PRECISION})"


@compiles(db.types.Float, "mysql")
def add_precision_to_mysql_floats(_element, _compiler, **_kw) -> str:
    """Forces floats to have minimum precision of 32, which converts the underlying type to be a
    double.  This is necessary because the default precision of floats is too low for some types,
    including unix timestamps, resulting in truncated values in MySQL.
    """
    return f"FLOAT({MYSQL_FLOAT_PRECISION})"


@compiles(db.types.FLOAT, "mysql")
def add_precision_to_mysql_FLOAT(_element, _compiler, **_kw) -> str:
    """Forces floats to have minimum precision of 32, which converts the underlying type to be a
    double.  This is necessary because the default precision of floats is too low for some types,
    including unix timestamps, resulting in truncated values in MySQL.
    """
    return f"FLOAT({MYSQL_FLOAT_PRECISION})"


class MySQLCompatabilityTypes:
    UniqueText = db.String(512)



import os
import sqlite3

import dagster._check as check


def create_db_conn_string(base_dir: str, db_name: str) -> str:
    check.str_param(base_dir, "base_dir")
    check.str_param(db_name, "db_name")

    path_components = os.path.abspath(base_dir).split(os.sep)
    db_file = f"{db_name}.db"
    return "sqlite:///{}".format("/".join(path_components + [db_file]))


def create_in_memory_conn_string(db_name: str) -> str:
    # Uses a named file-based URL for the in-memory connection (as opposed to the :memory: url) so
    # that multiple instances can share the same logical DB across connections, while maintaining
    # separate DBs for different db names.  The latter is required to have both the run / event_log
    # in-memory implementations within the same process
    return f"sqlite:///file:{db_name}?mode=memory&uri=true&cache=shared"


def get_sqlite_version() -> str:
    return str(sqlite3.sqlite_version)



from functools import update_wrapper
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Union,
    cast,
    overload,
)

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import public
from dagster._core.decorator_utils import format_docstring_for_description
from dagster._core.definitions.config import is_callable_valid_config_arg
from dagster._core.definitions.configurable import AnonymousConfigurableDefinition
from dagster._core.errors import DagsterInvalidDefinitionError, DagsterInvalidInvocationError
from dagster._utils import IHasInternalInit
from dagster._utils.backcompat import experimental_arg_warning

from ..decorator_utils import (
    get_function_params,
    has_at_least_one_parameter,
    is_required_param,
    positional_arg_name_list,
    validate_expected_params,
)
from .definition_config_schema import (
    CoercableToConfigSchema,
    IDefinitionConfigSchema,
    convert_user_facing_definition_config_schema,
)
from .resource_invocation import resource_invocation_result
from .resource_requirement import (
    RequiresResources,
    ResourceDependencyRequirement,
    ResourceRequirement,
)
from .scoped_resources_builder import (  # re-exported
    IContainsGenerator as IContainsGenerator,
    Resources as Resources,
    ScopedResourcesBuilder as ScopedResourcesBuilder,
)

if TYPE_CHECKING:
    from dagster._core.execution.resources_init import InitResourceContext

ResourceFunctionWithContext: TypeAlias = Callable[["InitResourceContext"], Any]
ResourceFunctionWithoutContext: TypeAlias = Callable[[], Any]
ResourceFunction: TypeAlias = Union[
    ResourceFunctionWithContext,
    ResourceFunctionWithoutContext,
]


class ResourceDefinition(AnonymousConfigurableDefinition, RequiresResources, IHasInternalInit):
    """Core class for defining resources.

    Resources are scoped ways to make external resources (like database connections) available to
    ops and assets during job execution and to clean up after execution resolves.

    If resource_fn yields once rather than returning (in the manner of functions decorable with
    :py:func:`@contextlib.contextmanager <python:contextlib.contextmanager>`) then the body of the
    function after the yield will be run after execution resolves, allowing users to write their
    own teardown/cleanup logic.

    Depending on your executor, resources may be instantiated and cleaned up more than once in a
    job execution.

    Args:
        resource_fn (Callable[[InitResourceContext], Any]): User-provided function to instantiate
            the resource, which will be made available to executions keyed on the
            ``context.resources`` object.
        config_schema (Optional[ConfigSchema): The schema for the config. If set, Dagster will check
            that config provided for the resource matches this schema and fail if it does not. If
            not set, Dagster will accept any config provided for the resource.
        description (Optional[str]): A human-readable description of the resource.
        required_resource_keys: (Optional[Set[str]]) Keys for the resources required by this
            resource. A DagsterInvariantViolationError will be raised during initialization if
            dependencies are cyclic.
        version (Optional[str]): (Experimental) The version of the resource's definition fn. Two
            wrapped resource functions should only have the same version if they produce the same
            resource definition when provided with the same inputs.
    """

    def __init__(
        self,
        resource_fn: ResourceFunction,
        config_schema: CoercableToConfigSchema = None,
        description: Optional[str] = None,
        required_resource_keys: Optional[AbstractSet[str]] = None,
        version: Optional[str] = None,
    ):
        self._resource_fn = check.callable_param(resource_fn, "resource_fn")
        self._config_schema = convert_user_facing_definition_config_schema(config_schema)
        self._description = check.opt_str_param(description, "description")
        self._required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys"
        )
        self._version = check.opt_str_param(version, "version")
        if version:
            experimental_arg_warning("version", "ResourceDefinition.__init__")

        # this attribute will be updated by the @dagster_maintained_resource and @dagster_maintained_io_manager decorators
        self._dagster_maintained = False

    @staticmethod
    def dagster_internal_init(
        *,
        resource_fn: ResourceFunction,
        config_schema: CoercableToConfigSchema,
        description: Optional[str],
        required_resource_keys: Optional[AbstractSet[str]],
        version: Optional[str],
    ) -> "ResourceDefinition":
        return ResourceDefinition(
            resource_fn=resource_fn,
            config_schema=config_schema,
            description=description,
            required_resource_keys=required_resource_keys,
            version=version,
        )

    @property
    def resource_fn(self) -> ResourceFunction:
        return self._resource_fn

    @property
    def config_schema(self) -> IDefinitionConfigSchema:
        return self._config_schema

    @public
    @property
    def description(self) -> Optional[str]:
        return self._description

    @public
    @property
    def version(self) -> Optional[str]:
        return self._version

    @public
    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        return self._required_resource_keys

    def _is_dagster_maintained(self) -> bool:
        return self._dagster_maintained

    @public
    @staticmethod
    def none_resource(description: Optional[str] = None) -> "ResourceDefinition":
        """A helper function that returns a none resource.

        Args:
            description ([Optional[str]]): The description of the resource. Defaults to None.

        Returns:
            [ResourceDefinition]: A resource that does nothing.
        """
        return ResourceDefinition.hardcoded_resource(value=None, description=description)

    @public
    @staticmethod
    def hardcoded_resource(value: Any, description: Optional[str] = None) -> "ResourceDefinition":
        """A helper function that creates a ``ResourceDefinition`` with a hardcoded object.

        Args:
            value (Any): The value that will be accessible via context.resources.resource_name.
            description ([Optional[str]]): The description of the resource. Defaults to None.

        Returns:
            [ResourceDefinition]: A hardcoded resource.
        """
        return ResourceDefinition(resource_fn=lambda _init_context: value, description=description)

    @public
    @staticmethod
    def mock_resource(description: Optional[str] = None) -> "ResourceDefinition":
        """A helper function that creates a ``ResourceDefinition`` which wraps a ``mock.MagicMock``.

        Args:
            description ([Optional[str]]): The description of the resource. Defaults to None.

        Returns:
            [ResourceDefinition]: A resource that creates the magic methods automatically and helps
                you mock existing resources.
        """
        from unittest import mock

        return ResourceDefinition(
            resource_fn=lambda _init_context: mock.MagicMock(), description=description
        )

    @public
    @staticmethod
    def string_resource(description: Optional[str] = None) -> "ResourceDefinition":
        return ResourceDefinition(
            resource_fn=lambda init_context: init_context.resource_config,
            config_schema=str,
            description=description,
        )

    def copy_for_configured(
        self,
        description: Optional[str],
        config_schema: CoercableToConfigSchema,
    ) -> "ResourceDefinition":
        resource_def = ResourceDefinition.dagster_internal_init(
            config_schema=config_schema,
            description=description or self.description,
            resource_fn=self.resource_fn,
            required_resource_keys=self.required_resource_keys,
            version=self.version,
        )

        resource_def._dagster_maintained = self._is_dagster_maintained()  # noqa: SLF001

        return resource_def

    def __call__(self, *args, **kwargs):
        from dagster._core.execution.context.init import UnboundInitResourceContext

        if has_at_least_one_parameter(self.resource_fn):
            if len(args) + len(kwargs) == 0:
                raise DagsterInvalidInvocationError(
                    "Resource initialization function has context argument, but no context was"
                    " provided when invoking."
                )
            if len(args) + len(kwargs) > 1:
                raise DagsterInvalidInvocationError(
                    "Initialization of resource received multiple arguments. Only a first "
                    "positional context parameter should be provided when invoking."
                )

            context_param_name = get_function_params(self.resource_fn)[0].name

            if args:
                check.opt_inst_param(args[0], context_param_name, UnboundInitResourceContext)
                return resource_invocation_result(
                    self, cast(Optional[UnboundInitResourceContext], args[0])
                )
            else:
                if context_param_name not in kwargs:
                    raise DagsterInvalidInvocationError(
                        f"Resource initialization expected argument '{context_param_name}'."
                    )
                check.opt_inst_param(
                    kwargs[context_param_name], context_param_name, UnboundInitResourceContext
                )

                return resource_invocation_result(
                    self, cast(Optional[UnboundInitResourceContext], kwargs[context_param_name])
                )
        elif len(args) + len(kwargs) > 0:
            raise DagsterInvalidInvocationError(
                "Attempted to invoke resource with argument, but underlying function has no context"
                " argument. Either specify a context argument on the resource function, or remove"
                " the passed-in argument."
            )
        else:
            return resource_invocation_result(self, None)

    def get_resource_requirements(
        self, outer_context: Optional[object] = None
    ) -> Iterator[ResourceRequirement]:
        source_key = cast(str, outer_context)
        for resource_key in sorted(list(self.required_resource_keys)):
            yield ResourceDependencyRequirement(key=resource_key, source_key=source_key)


def dagster_maintained_resource(
    resource_def: ResourceDefinition,
) -> ResourceDefinition:
    resource_def._dagster_maintained = True  # noqa: SLF001
    return resource_def


class _ResourceDecoratorCallable:
    def __init__(
        self,
        config_schema: Optional[Mapping[str, Any]] = None,
        description: Optional[str] = None,
        required_resource_keys: Optional[AbstractSet[str]] = None,
        version: Optional[str] = None,
    ):
        self.config_schema = config_schema  # checked by underlying definition
        self.description = check.opt_str_param(description, "description")
        self.version = check.opt_str_param(version, "version")
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys"
        )

    def __call__(self, resource_fn: ResourceFunction) -> ResourceDefinition:
        check.callable_param(resource_fn, "resource_fn")

        any_name = ["*"] if has_at_least_one_parameter(resource_fn) else []

        params = get_function_params(resource_fn)

        missing_positional = validate_expected_params(params, any_name)
        if missing_positional:
            raise DagsterInvalidDefinitionError(
                f"@resource decorated function '{resource_fn.__name__}' expects a single "
                "positional argument."
            )

        extras = params[len(any_name) :]

        required_extras = list(filter(is_required_param, extras))
        if required_extras:
            raise DagsterInvalidDefinitionError(
                f"@resource decorated function '{resource_fn.__name__}' expects only a single"
                " positional required argument. Got required extra params"
                f" {', '.join(positional_arg_name_list(required_extras))}"
            )

        resource_def = ResourceDefinition.dagster_internal_init(
            resource_fn=resource_fn,
            config_schema=self.config_schema,
            description=self.description or format_docstring_for_description(resource_fn),
            version=self.version,
            required_resource_keys=self.required_resource_keys,
        )

        # `update_wrapper` typing cannot currently handle a Union of Callables correctly
        update_wrapper(resource_def, wrapped=resource_fn)  # type: ignore

        return resource_def


@overload
def resource(config_schema: ResourceFunction) -> ResourceDefinition:
    ...


@overload
def resource(
    config_schema: CoercableToConfigSchema = ...,
    description: Optional[str] = ...,
    required_resource_keys: Optional[AbstractSet[str]] = ...,
    version: Optional[str] = ...,
) -> Callable[[ResourceFunction], "ResourceDefinition"]:
    ...


def resource(
    config_schema: Union[ResourceFunction, CoercableToConfigSchema] = None,
    description: Optional[str] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
    version: Optional[str] = None,
) -> Union[Callable[[ResourceFunction], "ResourceDefinition"], "ResourceDefinition"]:
    """Define a resource.

    The decorated function should accept an :py:class:`InitResourceContext` and return an instance of
    the resource. This function will become the ``resource_fn`` of an underlying
    :py:class:`ResourceDefinition`.

    If the decorated function yields once rather than returning (in the manner of functions
    decorable with :py:func:`@contextlib.contextmanager <python:contextlib.contextmanager>`) then
    the body of the function after the yield will be run after execution resolves, allowing users
    to write their own teardown/cleanup logic.

    Args:
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.resource_config`. If not set, Dagster will accept any config provided.
        description(Optional[str]): A human-readable description of the resource.
        version (Optional[str]): (Experimental) The version of a resource function. Two wrapped
            resource functions should only have the same version if they produce the same resource
            definition when provided with the same inputs.
        required_resource_keys (Optional[Set[str]]): Keys for the resources required by this resource.
    """
    # This case is for when decorator is used bare, without arguments.
    # E.g. @resource versus @resource()
    if callable(config_schema) and not is_callable_valid_config_arg(config_schema):
        return _ResourceDecoratorCallable()(config_schema)

    def _wrap(resource_fn: ResourceFunction) -> "ResourceDefinition":
        return _ResourceDecoratorCallable(
            config_schema=cast(Optional[Dict[str, Any]], config_schema),
            description=description,
            required_resource_keys=required_resource_keys,
            version=version,
        )(resource_fn)

    return _wrap


def make_values_resource(**kwargs: Any) -> ResourceDefinition:
    """A helper function that creates a ``ResourceDefinition`` to take in user-defined values.

        This is useful for sharing values between ops.

    Args:
        **kwargs: Arbitrary keyword arguments that will be passed to the config schema of the
            returned resource definition. If not set, Dagster will accept any config provided for
            the resource.

    For example:

    .. code-block:: python

        @op(required_resource_keys={"globals"})
        def my_op(context):
            print(context.resources.globals["my_str_var"])

        @job(resource_defs={"globals": make_values_resource(my_str_var=str, my_int_var=int)})
        def my_job():
            my_op()

    Returns:
        ResourceDefinition: A resource that passes in user-defined values.
    """
    return ResourceDefinition(
        resource_fn=lambda init_context: init_context.resource_config,
        config_schema=kwargs or Any,
    )
